from __future__ import annotations

import asyncio
import contextlib
import functools
import itertools
import json
import logging
import numbers
import os
import platform
import random
import re
import ssl
import string
import subprocess
import sys
import tempfile
import threading
import time
import traceback
import uuid
import warnings
from copy import deepcopy
from datetime import datetime, timedelta, timezone
from hashlib import md5
from logging.config import dictConfig
from math import ceil
from pathlib import Path
from typing import Dict, Iterable, List, NoReturn, Optional, Set, Tuple, Union
from urllib.parse import unquote, urlencode, urlparse
from zipfile import PyZipFile

import backoff
from packaging.version import Version

from coiled.context import get_datadog_trace_link, track_context

if sys.version_info >= (3, 8):
    from typing import Literal, Type, TypedDict
else:
    from typing_extensions import Literal, Type, TypedDict

import aiohttp
import boto3
import click
import dask.config
import dask.utils
import rich
import urllib3
import yaml
from dask.distributed import Security
from rich.console import Console

from coiled.exceptions import (
    AccountFormatError,
    ApiResponseStatusError,
    AuthenticationError,
    CidrInvalidError,
    GPUTypeError,
    InstanceTypeError,
    ParseIdentifierError,
    PortValidationError,
    UnsupportedBackendError,
)
from coiled.types import AWSOptions, FirewallOptions, GCPOptions

from .compatibility import COILED_VERSION, PY_VERSION
from .errors import ServerError

logger = logging.getLogger(__name__)

COILED_DIR = str(Path(__file__).parent)
ACCOUNT_REGEX = re.compile(r"^[a-z0-9]+(?:[-_][a-z0-9]+)*$")
ALLOWED_PROVIDERS = ["aws", "vm_aws", "gcp", "vm_gcp"]

COILED_LOGGER_NAME = "coiled"
COILED_SERVER = "https://cloud.coiled.io"
COILED_RUNTIME_REGEX = re.compile(r"^coiled/coiled-runtime-(?P<version>\d+\-\d+\-\d+)-*")

# Dots after family names ensure we do not match t3a or other variants
AWS_BALANCED_INSTANCES_FAMILY_FILTER = ("t2.", "t3.", "m5.", "m6i.", "t4g.", "m6g.", "m7g.")
AWS_INSTANCES_FAMILY_FILTER = AWS_BALANCED_INSTANCES_FAMILY_FILTER  # deprecated
AWS_GPU_INSTANCE_FAMILIES_FILTER = ("g4dn",)
AWS_UNBALANCED_INSTANCE_FAMILIES_FILTER = ("c5.", "c6i.", "r6i.", "c7g.", "r7g.")

GCP_SCHEDULER_GPU = {
    "scheduler_accelerator_type": "nvidia-tesla-t4",
    "scheduler_accelerator_count": 1,
}

# Directories to ignore when building wheels from source
IGNORE_PYTHON_DIRS = {"build", "dist", "docs", "tests"}

COILED_LOCAL_PACKAGE_PREFIX = "coiled_local_"


class VmType(TypedDict):
    """
    Example:
        {
        'name': 't2d-standard-8',
        'cores': 8,
        'gpus': 0,
        'gpu_name': None,
        'memory': 32768,
        'backend_type': 'vm_gcp'
        }
    """

    name: str
    cores: int
    gpus: int
    gpu_name: str
    memory: int
    backend_type: str
    coiled_credits: float


# TODO: copied from distributed, introduced in 2021.12.0.
# We should be able to remove this someday once we can increase
# the minimum supported version.
def in_async_call(loop, default=False):
    """Whether this call is currently within an async call"""
    try:
        return loop.asyncio_loop is asyncio.get_running_loop()
    except RuntimeError:
        # No *running* loop in thread. If the event loop isn't running, it
        # _could_ be started later in this thread though. Return the default.
        if not loop.asyncio_loop.is_running():
            return default
        return False


def validate_account(account: str):
    if ACCOUNT_REGEX.match(account) is None:
        raise AccountFormatError(
            f"Bad account format. Account '{account}' should be a combination "
            "of lowercase letters, numbers and hyphens."
        )


def random_str(length: int = 8):
    return "".join(random.choices(string.ascii_lowercase + string.digits, k=length))


class GatewaySecurity(Security):
    """A security implementation that temporarily stores credentials on disk.

    The normal ``Security`` class assumes credentials already exist on disk,
    but our credentials exist only in memory. Since Python's SSLContext doesn't
    support directly loading credentials from memory, we write them temporarily
    to disk when creating the context, then delete them immediately."""

    def __init__(self, tls_key, tls_cert, extra_conn_args: Optional[dict] = None):
        self.tls_scheduler_key = tls_key
        self.tls_scheduler_cert = tls_cert
        self.extra_conn_args = extra_conn_args or {}

    def __repr__(self):
        return "GatewaySecurity<...>"

    def get_connection_args(self, role):
        ctx = None
        if self.tls_scheduler_key and self.tls_scheduler_cert:
            with tempfile.TemporaryDirectory() as tempdir:
                key_path = os.path.join(tempdir, "dask.pem")
                cert_path = os.path.join(tempdir, "dask.crt")
                with open(key_path, "w") as f:
                    f.write(self.tls_scheduler_key)
                with open(cert_path, "w") as f:
                    f.write(self.tls_scheduler_cert)
                ctx = ssl.create_default_context(purpose=ssl.Purpose.SERVER_AUTH, cafile=cert_path)
                ctx.verify_mode = ssl.CERT_REQUIRED
                ctx.check_hostname = False
                ctx.load_cert_chain(cert_path, key_path)
        return {
            "ssl_context": ctx,
            "require_encryption": True,
            "extra_conn_args": self.extra_conn_args,
        }


async def handle_api_exception(response, exception_cls: Type[Exception] = ServerError) -> NoReturn:
    with contextlib.suppress(aiohttp.ContentTypeError, json.JSONDecodeError, AttributeError):
        # First see if it's an error we created that has a more useful
        # message
        error_body = await response.json()
        if isinstance(error_body, dict):
            errs = error_body.get("non_field_errors")
            if "message" in error_body:
                raise exception_cls(error_body["message"])
            elif errs and isinstance(errs, list) and len(errs):
                raise exception_cls(errs[0])
            else:
                raise exception_cls("Server error, ".join(f"{k}={v}" for (k, v) in error_body.items()))
        else:
            raise exception_cls(error_body)

    error_text = await response.text()

    if not error_text:
        # Response contains no text/body, let's not raise an empty exception
        error_text = f"{response.status} - {response.reason}"
    raise exception_cls(error_text)


def normalize_server(server: Optional[str]) -> str:
    if not server:
        server = COILED_SERVER
    # Check if using an older server
    if "beta.coiledhq.com" in server or "beta.coiled.io" in server:
        # NOTE: https is needed here as http is not redirecting
        server = COILED_SERVER

    # remove any trailing slashes
    server = server.rstrip("/")

    return server


def get_account_membership(user_dict: dict, memberships: list, account: Optional["str"] = None) -> Optional[dict]:
    if account is None:
        account = user_dict.get("default_account")
        account = account.get("slug") if account else None

    for membership in memberships:
        account_details = membership.get("account", {})
        has_membership = account_details.get("slug") == account
        if has_membership:
            return membership

    else:
        return None


def get_auth_header_value(token: str) -> str:
    """..."""
    # TODO: delete the branching after client only supports ApiToken.
    if "-" in token:
        return "ApiToken " + token
    else:
        return "Token " + token


def has_program_quota(account_usage: dict) -> bool:
    return account_usage.get("has_quota") is True


async def login_if_required(
    *,
    server: Optional[str] = None,
    token: Optional[str] = None,
    account: Optional[str] = None,
    save: Optional[bool] = None,
    use_config: bool = True,
    retry: bool = True,
    browser: Optional[bool] = None,
):
    account = account or dask.config.get("coiled.account", None)
    # "save" bool means always/never, None means try to do the thing that makes sense
    if save is None:
        # if token is already set in config and isn't being changed, then no need to save again
        if dask.config.get("coiled.token", None) and (not token or token == dask.config.get("coiled.token")):
            save = False
        else:
            # user doesn't already have token saved, so save it without asking
            save = True

    if use_config:
        token = token or dask.config.get("coiled.token")
        server = server or dask.config.get("coiled.server")
        if "://" not in server:
            server = "http://" + server
        server = server.rstrip("/")
        account = account or dask.config.get("coiled.account")

    try:
        await handle_credentials(server=server, token=token, account=account, save=save, retry=retry, browser=browser)
    except ImportError as e:
        rich.print(f"[red]{e}")


@backoff.on_exception(backoff.expo, ApiResponseStatusError, logger=logger)
async def _fetch_data(*, session, server, endpoint):
    response = await session.request("GET", f"{server}{endpoint}")
    if response.status == 426:
        # client version upgrade required
        await handle_api_exception(response)
    elif response.status in [401, 403]:
        raise AuthenticationError(f"Auth Error: {response.status}. Invalid Token")
    elif response.status in [502, 503, 504]:
        raise ApiResponseStatusError(
            f"Unable to receive data from the server. Received {response.status} - Temporary Error."
        )
    elif response.status >= 400:
        await handle_api_exception(response)
    return await response.json()


# We only use ApiResponseStatusError on status 502, 503, 504
async def handle_credentials(
    *,
    server: Optional[str] = None,
    token: Optional[str] = None,
    account: Optional[str] = None,
    save: Optional[bool] = None,
    retry: bool = True,
    print_invalid_token_messages: bool = True,
    browser: Optional[bool] = None,
) -> Tuple[str, str, str, list]:
    """Validate and optionally save credentials

    Parameters
    ----------
    server
        Server to connect to. If not specified, will check the
        ``coiled.server`` configuration value.
    token
        Coiled user token to use. If not specified, will prompt user
        to input token.
    account
        Name of Coiled account to use. If not provided, will
        default to the user's default account.
    save
        Whether or not save credentials to coiled config file.
        If ``None``, will ask for input on whether or not credentials
        should be saved. Defaults to None.
    retry
        Whether or not to try again if invalid credentials are entered.
        Retrying is often desired in interactive situations, but not
        in more programmatic scenerios. Defaults to True.

    Returns
    -------
    user
        Username
    token
        User API token
    server
        Server being used
    memberships
        List of account memberships
    """

    is_account_specified = bool(account)
    if account:
        validate_account(account)

    # If testing locally with `ngrok` we need to
    # rewrite the server to localhost
    server = server or dask.config.get("coiled.server", COILED_SERVER)
    server = normalize_server(server)

    browser = True if browser is None else browser

    if token is None:
        from .auth import client_token_grant_flow

        result = await client_token_grant_flow(server, browser, account=account)
        if result:
            return result
        raise ValueError(
            "Authorization failed. Please try to login again, and if the error persists, "
            "please reach out to Coiled Support at support@coiled.io"
        )

    # TODO: revert when we remove versioneer
    if dask.config.get("coiled.no-minimum-version-check", False):
        client_version = "coiled-frontend-js"
    else:
        client_version = COILED_VERSION
    account_usage = {}
    # Validate token and get username
    async with aiohttp.ClientSession(
        headers={
            "Authorization": get_auth_header_value(token),
            "Client-Version": client_version,
        }
    ) as session:
        try:
            user_dict = await _fetch_data(
                session=session,
                server=server,
                endpoint="/api/v2/user/me",
            )
        except AuthenticationError:
            if print_invalid_token_messages:
                rich.print("[red]Invalid Coiled token encountered[/red]")
            if retry:
                return await handle_credentials(server=server, token=None, account=account, save=None, retry=False)
            else:
                if print_invalid_token_messages:
                    rich.print(
                        "You can use [green]coiled login[/green] to authorize a new token for your Coiled client, "
                        "or contact us at support@coiled.io if you continue to have problems."
                    )
                raise
        memberships = await _fetch_data(session=session, server=server, endpoint="/api/v2/user/me/memberships")
        if not isinstance(memberships, list) or not memberships:
            account_membership = None
            memberships = []
        else:
            account_membership = get_account_membership(user_dict, memberships, account=account)
        # only validate if account arg is provided by user
        if account and not account_membership:
            rich.print("[red]You are not a member of this account. Perhaps try another one?\n")
            account = click.prompt("Account")
            if account:
                validate_account(account)
            else:
                rich.print("[red]No account provided, unable to login.")

            return await handle_credentials(server=server, token=token, save=None, account=account)

        if account_membership and not account:
            account = account_membership.get("account", {}).get("slug")

        # We should always have username from above, but let's be defensive about it just in case.
        user = user_dict.get("username")
        if not user:
            raise ValueError(
                "Unable to get your account username after login. Please try to login again, if "
                "the error persists, please reach out to Coiled Support at support@coiled.io"
            )

        # Avoid extra printing when creating clusters
        if save is not False:
            rich.print("[green]Authentication successful[/green] 🎉")
            # Only get account usage when we are actually going to check it
            account_usage = await _fetch_data(
                session=session, server=server, endpoint=f"/api/v2/user/account/{account}/usage"
            )
            if not isinstance(account_usage, dict):
                account_usage = {}
            if not has_program_quota(account_usage):
                rich.print("[red]You have reached your quota of Coiled credits for this account.")
    if save is None:
        # Optionally save user credentials for next time
        save_creds = input("Save credentials for next time? [Y/n]: ")
        if save_creds.lower() in ("y", "yes", ""):
            save = True
    if save:
        creds = {
            "coiled": {
                "user": user,
                "token": token,
                "server": server,
            }
        }
        if is_account_specified and account:
            # If user didn't specify account, don't set it in the local config file because
            # when it's set, it overrides default account set on server.
            creds["coiled"]["account"] = account

        config, config_file = save_config(new_config=creds)
        rich.print(f"Credentials have been saved at {config_file}")
        # Make sure saved configuration values are set for the current Python process.
        dask.config.update(dask.config.config, config)

    return user, token, server, memberships


def save_config(new_config: dict) -> tuple[dict, str]:
    """Save new config values to Coiled config file, return new config and path to file."""
    config_file = os.path.join(os.path.expanduser("~"), ".config", "dask", "coiled.yaml")
    # Make sure directory with config exists
    os.makedirs(os.path.dirname(config_file), exist_ok=True)
    configs = dask.config.collect_yaml([config_file])

    config = dask.config.merge(*configs, new_config)
    with open(config_file, "w") as f:
        f.write(yaml.dump(config))
    return config, config_file


class Spinner:
    """A spinner context manager to denotate we are still working"""

    def __init__(self, delay=0.2):
        self.spinner = itertools.cycle(["-", "/", "|", "\\"])
        self.delay = delay
        self.busy = False

    def write_next(self):
        with self._screen_lock:
            sys.stdout.write(next(self.spinner))
            sys.stdout.flush()

    def remove_spinner(self, cleanup=False):
        with self._screen_lock:
            sys.stdout.write("\b")
            if cleanup:
                sys.stdout.write(" ")  # overwrite spinner with blank
                sys.stdout.write("\r")  # move to next line
            sys.stdout.flush()

    def spinner_task(self):
        while self.busy:
            self.write_next()
            time.sleep(self.delay)
            self.remove_spinner()

    def __enter__(self):
        if sys.stdout.isatty():
            self._screen_lock = threading.Lock()
            self.busy = True
            self.thread = threading.Thread(target=self.spinner_task)
            self.thread.start()

    def __exit__(self, exception, value, tb):
        if sys.stdout.isatty():
            self.busy = False
            self.remove_spinner(cleanup=True)
        else:
            sys.stdout.write("\r")


def parse_identifier(
    identifier: str,
    property_name: str = "name",
    can_have_revision: bool = False,
    allow_uppercase: bool = False,
) -> Tuple[Optional[str], str, Optional[str]]:
    """
    Parameters
    ----------
    identifier:
        identifier of the resource, i.e. "coiled/xgboost" "coiled/xgboost:1ef489", "xgboost:1ef489" or "xgboost"
    can_have_revision:
        Indicates if this identifier supports having a ":<string>" postfix, as in
        the ":1ef489" in "xgboost:1ef489". At time of writing, this only has an effect
        on software environments. For other resources this has no meaning. At time
        of writing, this only affects the error message that will be printed out.
    property_name:
        The name of the parameter that was being validated; will be printed
        with any error messages, i.e. Unsupported value for "software_environment".

    Examples
    --------
    >>> parse_identifier("coiled/xgboost", "software_environment")
    ("coiled", "xgboost", None)
    >>> parse_identifier("xgboost", "software_environment", False)
    (None, "xgboost", None)
    >>> parse_identifier("coiled/xgboost:1ef4543", "software_environment", True)
    ("coiled", "xgboost", "1ef4543")

    Raises
    ------
    ParseIdentifierError
    """
    if allow_uppercase:
        rule = re.compile(r"^([a-zA-Z0-9-_]+?/)?([a-zA-Z0-9-_]+?)(:[a-zA-Z0-9-_]+)?$")
        rule_text = ""
    else:
        rule = re.compile(r"^([a-z0-9-_]+?/)?([a-z0-9-_]+?)(:[a-zA-Z0-9-_]+)?$")
        rule_text = "lowercase "

    match = re.fullmatch(rule, identifier)
    if match:
        account, name, revision = match.groups()
        account = account.replace("/", "") if account else account
        revision = revision.replace(":", "") if revision else revision
        return account, name, revision

    if can_have_revision:
        message = (
            f"'{identifier}' is invalid {property_name}: should have format (<account>/)<name>(:<revision>),"
            ' for example "coiled/xgboost:1efd456", "xgboost:1efd456" or "xgboost". '
            f"It can only contain {rule_text}ASCII letters, numbers, hyphens and underscores. "
            "The <name> is required (xgboost in the previous example). The <account> prefix"
            f" can be used to specify a {property_name} from a different account, and the"
            f" :<revision> can be used to select a specific revision of the {property_name}."
        )
    else:
        message = (
            f"'{identifier}' is invalid {property_name}: should have format (<account>/)<name>,"
            f' for example "coiled/xgboost" or "python-37". and can only contain {rule_text}ASCII letters,'
            ' numbers, hyphens and underscores. The <name> is required ("xgboost" and "python-37"'
            " in the previous examples)."
            f' The <account> prefix (i.e. "coiled/") can be used to specify a {property_name}'
            " from a different account."
        )
    raise ParseIdentifierError(message)


def get_platform():
    # Determine platform
    if sys.platform == "linux":
        platform = "linux"
    elif sys.platform == "darwin":
        platform = "osx"
    elif sys.platform == "win32":
        platform = "windows"
    else:
        raise ValueError(f"Invalid platform {sys.platform} encountered")
    return platform


class ExperimentalFeatureWarning(RuntimeWarning):
    """Warning raise by an experimental feature"""

    pass


class DeprecatedFeatureWarning(RuntimeWarning):
    pass


def experimental(func):
    @functools.wraps(func)
    def inner(*args, **kwargs):
        warnings.warn(
            f"{func.__name__} is an experimental feature which is subject "
            "to breaking changes, being removed, or otherwise updated without notice "
            "and should be used accordingly.",
            ExperimentalFeatureWarning,
        )
        return func(*args, **kwargs)

    return inner


async def run_command_in_subprocess(cmd: str, suppress_error: bool = False):
    """Run the command in subprocess.

    If a user is using a python version < 3.8 then he will get a RuntimeError exception -
    ``Cannot add child handler, the child watcher does not have a loop attached``. As a
    workaround for the issue, we check if the user is running any earlier version of python
    and run the command in subprocess instead.

    For reference on the RuntimeError exception: https://bugs.python.org/issue35621

    """
    platform = get_platform()

    if PY_VERSION < Version("3.8") or platform == "windows":
        proc = subprocess.Popen(
            cmd.split(" "),
            bufsize=1,
            universal_newlines=True,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
        )
        assert proc.stdout
        for line in proc.stdout:
            yield line.rstrip()

    else:
        proc = await asyncio.create_subprocess_shell(
            cmd,
            stdout=asyncio.subprocess.PIPE,
            stderr=asyncio.subprocess.PIPE,
        )

        lines = []
        while proc.stdout and proc.stderr and proc.returncode is None:
            await asyncio.sleep(0.5)
            async for line in proc.stdout:
                line = line.decode(encoding=sys.stdout.encoding or "utf-8").rstrip()
                lines.append(line)
                yield line

            async for line in proc.stderr:
                line = line.decode(encoding=sys.stderr.encoding or "utf-8").rstrip()
                lines.append(line)
                yield line

        if proc.returncode and not suppress_error:
            raise ValueError("\n".join(lines))


def name_exists_in_dict(user: Union[str, None], name: Union[str, None], dictionary: Union[dict, None]) -> bool:
    """Check if name exists in dictionary.

    Since we are returning ``<account>/<name>`` when listing software envs
    and cluster configs, this method will try to handle the case where sometimes
    we are passing a name with the format ``<account>/<name>`` and sometimes with
    ``<name``.

    If we can't find ``<account>/<name>`` in the dictionary, we will check if we
    can find the ``name`` substring in the dictionary keys.

    """
    if dictionary:
        if f"{user}/{name}" in dictionary or name in dictionary:
            return True
        if any([soft_env for soft_env in dictionary if name in soft_env]):
            return True
    return False


def rich_console():
    is_spyder = False

    with contextlib.suppress(AttributeError, ImportError):
        from IPython.core.getipython import get_ipython

        ipython = get_ipython()
        if ipython and ipython.config.get("SpyderKernelApp"):
            is_spyder = True

    if is_spyder:
        print("Creating Cluster. This usually takes 1-2 minutes...")
        return Console(force_jupyter=False)
    return Console()


@backoff.on_exception(backoff.expo, Exception, max_time=10)
def verify_aws_credentials_with_retry(aws_access_key_id: str, aws_secret_access_key: str):
    verify_aws_credentials(aws_access_key_id=aws_access_key_id, aws_secret_access_key=aws_secret_access_key)


def verify_aws_credentials(aws_access_key_id: str, aws_secret_access_key: str):
    """Verify AWS Credentials are valid, raise exception so caller knows what is wrong with credentials."""
    sts = boto3.client(
        "sts",
        aws_access_key_id=aws_access_key_id,
        aws_secret_access_key=aws_secret_access_key,
    )
    sts.get_caller_identity()


def scheduler_ports(protocol: Union[str, List[str]]):
    """Generate scheduler ports based on protocol(s)"""
    exclude = 8787  # dashboard port
    start = 8786
    if isinstance(protocol, str):
        return start

    port = start
    ports = []
    for _ in protocol:
        if port == exclude:
            port += 1
        ports.append(port)
        port += 1
    return ports


def parse_gcp_region_zone(region: Optional[str] = None, zone: Optional[str] = None):
    """Parse GCP zone and region or return default region/zone.

    This is an helper function to make it easier for us
    to parse gcp zones. Since users can specify regions,
    zones or one of the two, we should create some sane
    way to deal with the different combinations.

    """
    if not region and not zone:
        region = "us-east1"
        zone = "us-east1-c"
    elif region and zone and len(zone) == 1:
        zone = f"{region}-{zone}"
    elif not region and zone and len(zone) == 1:
        region = "us-east1"
        zone = f"{region}-{zone}"
    elif zone and not region:
        region = zone[:-2]

    return region, zone


class UTCFormatter(logging.Formatter):
    converter = time.gmtime


def enable_debug_mode():
    from .context import COILED_SESSION_ID

    LOGGING = {
        "version": 1,
        "disable_existing_loggers": True,
        "formatters": {
            "utc": {
                "()": UTCFormatter,
                "format": "[UTC][%(asctime)s][%(levelname)-8s][%(name)s] %(message)s",
            },
        },
        "handlers": {
            "utc-console": {
                "class": "logging.StreamHandler",
                "formatter": "utc",
            },
        },
        "loggers": {
            "coiled": {
                "handlers": ["utc-console"],
                "level": logging.DEBUG,
            },
        },
    }
    dictConfig(LOGGING)
    logger.debug(f"Coiled Version : {COILED_VERSION}")
    start = datetime.now(tz=timezone.utc)
    trace_link = get_datadog_trace_link(start=start, **{"coiled-session-id": COILED_SESSION_ID})
    logger.info(f"DD Trace: {trace_link}")
    trace_link = get_datadog_logs_link(start=datetime.now(tz=timezone.utc), **{"coiled-session-id": COILED_SESSION_ID})
    logger.info(f"DD Logs: {trace_link}")


def get_datadog_logs_link(
    start: Optional[datetime] = None,
    end: Optional[datetime] = None,
    **filters: Dict[str, str],
):
    params = {
        "query": " ".join([f"@{k}:{v}" for k, v in filters.items()]),
        "live": "false",
    }
    if start:
        fuzzed = start - timedelta(minutes=1)
        params["from_ts"] = str(int(fuzzed.timestamp() * 1000))
    if end:
        fuzzed = end + timedelta(minutes=1)
        params["to_ts"] = str(int(fuzzed.timestamp() * 1000))
    return f"https://app.datadoghq.com/logs?{urlencode(params)}"


def validate_gpu_type(gpu_type: str):
    """Validate gpu type provided by the user.

    Currently, we are just accepting the nvidia-tesla-t4 gpu type
    but in the next iteration we will accept more types. We are also
    filtering out virtual workstation gpus upstream, so we need this
    function to remove types that we know it will fail.

    """

    if gpu_type != "nvidia-tesla-t4":
        error_msg = f"GPU type '{gpu_type}' is not a supported GPU type. Allowed " "GPU types are: 'nvidia-tesla-t4'."
        raise GPUTypeError(error_msg)
    return True


def is_gcp(account: str, accounts: dict) -> bool:
    """Determine if an account backend is Google Cloud Provider.

    Parameters
    ----------
    account: str
        Slug of Coiled account to use.
    accounts: dict
        Dictionary of available accounts from the /api/v1/users/me endpoint.

    Returns
    -------
    gcp_backend: bool
        True if a GCP backend, else False.
    """
    user_account = accounts.get(account, {})
    user_options = user_account.get("options", {})
    gcp_backend = user_account.get("backend") == "vm_gcp" or user_options.get("provider_name") == "gcp"
    return gcp_backend


def is_customer_hosted(account: str, accounts: dict) -> bool:
    """Determine if an account backend is customer-hosted.

    Parameters
    ----------
    account: str
        Slug of Coiled account to use.
    accounts: dict
        Dictionary of available accounts from the /api/v1/users/me endpoint.

    Returns
    -------
    customer_hosted: bool
        True if a customer-hosted backend, else False.
    """
    user_account = accounts.get(account, {})
    backend_type = user_account.get("backend")
    customer_hosted = backend_type == "vm"
    return customer_hosted


def validate_cidr_block(cidr_block: str):
    """Validate cidr block added by the user.

    Here we are only checking if the cidr block is in the
    format <int>.<int>.<int>.<int>/<int>.

    """
    if not isinstance(cidr_block, str):
        raise CidrInvalidError(
            f"CIDR needs to be of type string, but received '{type(cidr_block)}' "
            "please specify your CIDR block as a string."
        )
    match = re.fullmatch(r"(\d{1,3}\.){3}\d{1,3}\/\d{1,3}", cidr_block)
    if not match:
        raise CidrInvalidError(
            f"The CIDR block provided '{cidr_block}' doesn't appear "
            "to have the correct IPV4 pattern, your CIDR block should "
            "follow the '0.0.0.0/0' pattern."
        )
    return True


def validate_ports(ports):
    """Validate the ports tha the user tries to pass.

    We need to make sure that the user passes a list of ints only,
    otherwise we should raise an exception.

    """
    if not isinstance(ports, list):
        raise PortValidationError(
            f"Ports need to be of type list, but received '{type(ports)}' " "please adds your ports in a list."
        )

    for port in ports:
        if not isinstance(port, int):
            raise PortValidationError(
                f"Ports need to be of type int, but received '{type(port)}' "
                "please use a int value for your port number."
            )
    return True


def validate_network(backend_options: dict, is_customer_hosted: bool):
    """Validate network configuration from backend options.

    Users can specify network and/or subnet(s) to use. We need to validate these and
    ensure that they're only used with customer hosted backend.
    """
    if not is_customer_hosted:
        raise UnsupportedBackendError(
            "Network configuration isn't available in a Coiled "
            "hosted backend. Please change your backend to configure network."
        )

    network_options = backend_options.get("network", {})

    # check that resource ids, if specified, are strings
    resource_ids = (
        network_options.get("network_id"),
        network_options.get("scheduler_subnet_id"),
        network_options.get("worker_subnet_id"),
        network_options.get("firewall_id"),
    )

    for resource_id in resource_ids:
        if resource_id and not isinstance(resource_id, str):
            raise TypeError(f"Network id '{id}' for '{resource_id}' must be a string.")


def parse_backend_options(
    backend_options: dict,
    account: str,
    accounts: dict,
    worker_gpu: Optional[int],
) -> dict:
    """Parse backend options before launching cluster.

    The following are checked and parsed before launching a cluster:
        - If launching into a GCP cluster, `preemptible` is aliased with `spot`.
        - If requesting worker GPUs, `spot` defaults to `False` unless specified.
        - If set, `zone` overrides `gcp_zone` (for GCP).

    Parameters
    ----------
    backend_options: dict
        Dictionary of backend specific options (e.g. ``{'region': 'us-east-2'}``).
    account: str
        Slug of Coiled account to use.
    accounts: dict
        Dictionary of available accounts from the /api/v1/users/me endpoint.
    worker_gpu: int
        Number of GPUs allocated for each worker.

    Returns
    -------
    backend_options: dict
        A dictionary of parsed backend options.
    """
    backend_options = deepcopy(backend_options)
    gcp_backend = is_gcp(account, accounts)
    customer_hosted = is_customer_hosted(account, accounts)
    # alias preemptible/spot
    if backend_options.get("preemptible") is not None and gcp_backend:
        backend_options["spot"] = backend_options.pop("preemptible")
    # default to on-demand for gpu workers
    if backend_options.get("spot") is None and bool(worker_gpu):
        backend_options["spot"] = False

    # zone (if set) overrides gcp_zone for gcp
    if backend_options.get("zone") and backend_options.get("gcp_project_name"):
        backend_options["gcp_zone"] = backend_options.get("zone")

    if backend_options.get("network"):
        validate_network(backend_options, customer_hosted)
    if backend_options.get("firewall"):
        cidr = backend_options["firewall"].get("cidr")
        ports = backend_options["firewall"].get("ports")
        if ports:
            validate_ports(ports)
        if cidr:
            validate_cidr_block(cidr)
    return backend_options


async def get_worker_creation_notification_error(cloud, account: Optional[str]) -> Optional[str]:
    """Call get_notifications and return error messages from worker creation.

    When we try to create workers, we use the ``create_worker`` event type to
    the notification, we can quickly use this event type and the logging level
    ``ERROR`` to only get notifications that are related to worker creation.

    Note that we don't use the cluster name in the backend, but the cluster id
    so we can't really filter the notifications by name, with that in mind, we
    make the assumption that the last notification that we receive will be the
    one from the cluster that just ran. This is ugly but should work okay for
    now.

    """
    notifications = await cloud.get_notifications(account=account, json=True, event_type="create_worker", level="ERROR")

    if notifications:
        last_notification = notifications[-1]
        error_message = last_notification.get("data", {}).get("error")

        # If there is no reason, lets avoiding showing "giving up" messages
        # with no useful message
        if error_message and "Reason" in error_message:
            return error_message
        return last_notification["msg"]


def parse_wait_for_workers(n_workers: int, wait_for_workers: Optional[bool | int | float] = None) -> int:
    """Parse the option to wait for workers."""
    wait_for_workers = (
        # Set 30% as default value to wait for workers
        dask.config.get("coiled.wait-for-workers", 0.3) if wait_for_workers is None else wait_for_workers
    )

    if wait_for_workers is True:
        to_wait = n_workers
    elif wait_for_workers is False:
        to_wait = 0
    elif isinstance(wait_for_workers, int):
        if wait_for_workers >= 0 and wait_for_workers <= n_workers:
            to_wait = wait_for_workers
        else:
            raise ValueError(
                f"Received invalid value '{wait_for_workers}' as wait_for_workers, "
                f"this value needs to be between 0 and {n_workers}"
            )
    elif isinstance(wait_for_workers, float):
        if wait_for_workers >= 0 and wait_for_workers <= 1.0:
            to_wait = ceil(wait_for_workers * n_workers)
        else:
            raise ValueError(
                f"Received invalid value '{wait_for_workers}' as wait_for_workers, "
                "this value needs to be a value between 0 and 1.0."
            )
    else:
        raise ValueError(
            f"Received invalid value '{wait_for_workers}' as wait_for_workers, "
            "this value needs to be either a Boolean, an Integer or a Float."
        )

    return to_wait


def bytes_to_mb(bytes: float) -> int:
    """Convert bytes to Mb.

    Our backend expects the memory to be passed as Mb, so we need to convert
    bytes obtained by dask to Mb.

    """
    return round(abs(bytes) * 10**-6)


def parse_requested_memory(
    memory: Optional[Union[int, str, float, List[str], List[int], List[float]]],
    min_memory: Optional[Union[str, int, float]],
) -> dict:
    """Handle memory requested by user.

    Users calling `list_instance_types` can use different ways to specify memory,
    this function will handle the different cases and return a dictionary with the
    expected result.

    Note: We are also giving a +- 10% buffer on the requested memory.

    """
    parsed_memory = {}

    if isinstance(memory, list):
        if len(memory) > 2:
            raise ValueError(
                f"Memory should contain only two values in the format `[minimum, maximum]`, but received {memory}."
            )
        parsed_memory["memory__gte"] = bytes_to_mb(dask.utils.parse_bytes(memory[0]) * 0.89)
        parsed_memory["memory__lte"] = bytes_to_mb(dask.utils.parse_bytes(memory[1]) * 1.1)
    elif isinstance(memory, int) or isinstance(memory, float) or isinstance(memory, str):
        parsed_memory["memory__gte"] = bytes_to_mb(dask.utils.parse_bytes(memory) * 0.89)
        parsed_memory["memory__lte"] = bytes_to_mb(dask.utils.parse_bytes(memory) * 1.1)

    if min_memory and not memory:
        parsed_memory["memory__gte"] = bytes_to_mb(dask.utils.parse_bytes(min_memory))

    return parsed_memory


@track_context
def get_instance_type_from_cpu_memory(
    cpu: Optional[Union[int, List[int]]] = None,
    memory: Optional[Union[int, str, float, List[int], List[str], List[float]]] = None,
    gpus: Optional[int] = None,
    backend: Optional[str] = None,
    arch: Literal["x86_64", "arm64"] = "x86_64",
) -> List[str]:
    """Get instances by cpu/memory combination.

    We are using the `list_instance_types` method to get the
    instances that match the cpu/memory specification. If no
    instance can be found we will raise an exception
    informing the user about this.

    """
    # Circular imports
    from .core import list_instance_types

    if backend == "gcp":
        instance_family = "n1" if gpus else "e2"
        instances = [
            instance_name
            for instance_name in list_instance_types(cores=cpu, memory=memory, backend=backend, arch=arch)
            if instance_name.startswith(instance_family)
        ]
    else:
        instance_types = list_instance_types(cores=cpu, memory=memory, backend=backend, gpus=gpus, arch=arch)

        aws_family_filter = AWS_BALANCED_INSTANCES_FAMILY_FILTER
        if gpus:
            aws_family_filter = AWS_GPU_INSTANCE_FAMILIES_FILTER
        elif cpu and memory:
            aws_family_filter += AWS_UNBALANCED_INSTANCE_FAMILIES_FILTER

        instances = [instance_name for instance_name in instance_types if instance_name.startswith(aws_family_filter)]

    if not instances:
        error_message = (
            "\n"
            "Unable to find instance types that match the specification: \n"
            f"    Cores: {cpu}  Memory: {memory} GPUs: {gpus}  Arch: {arch}\n"
        )

        if (isinstance(memory, numbers.Number) and memory < 1e9) or (  # type: ignore
            isinstance(memory, (tuple, list)) and all(isinstance(m, numbers.Number) and m < 1e9 for m in memory)  # type: ignore
        ):
            error_message += f"""
Your specified memory, {memory}, was too low. Memory units are in bytes.
We recommend strings like "32 GiB", or numbers like 32e9.
"""

        if cpu or memory:
            error_message += (
                "\n"
                "You can select a range for the cpu or memory, for example:"
                "\n    `cpu=[2, 8]` or `memory=['32 GiB','64GiB'] \n"
            )

        if cpu and memory:
            memory_matching_instances = [
                instance_name for instance_name in list_instance_types(memory=memory, backend=backend)
            ]

            if memory_matching_instances:
                error_message += (
                    "You might want to pick these instances that match your "
                    "memory requirements, but have different core count. \n"
                    f"{memory_matching_instances} \n"
                )
        error_message += (
            "\n"
            "You can also use the following to list instance types:"
            f'\n    coiled.list_instance_types(backend="{backend}")\n'
            "\nand use the `scheduler_vm_types=[]` or `worker_vm_types=[]`"
            "\nkeyword arguments to specify your desired instance type."
        )
        raise InstanceTypeError(error_message)
    return instances


def any_gpu_instance_type(type_list: Optional[Union[str, List[str]]]) -> bool:
    # TODO (future PR) ideally we'd get this from database via endpoint as part of preflight check
    #   but for now we'll just use regex to match instance types with bundled GPU
    if not type_list:
        return False
    if isinstance(type_list, str):
        type_list = [type_list]
    # check supported AWS instance types and GCP type with non-guest GPU (A100)
    return any(
        re.search(r"^(((p4|p3|g5|g5g|g4dn|g4ad)\.)|(a2-))", vm_type, flags=re.IGNORECASE) for vm_type in type_list
    )


def validate_backend_options(backend_options):
    # both typing.TypedDict and typing_extensions.TypedDict have __optional_keys__
    aws_keys = set(AWSOptions.__optional_keys__)  # type: ignore
    gcp_keys = set(GCPOptions.__optional_keys__)  # type: ignore

    # show warning for unknown keys
    all_keys = set().union(aws_keys, gcp_keys)
    for key in backend_options.keys():
        if key in ("firewall_spec",):
            # firewall_spec isn't currently in user-schema, but it's how we send the data
            # to backend endpoint.
            # `{"ingress": [...]}` (user) —> `{firewall_spec: {"ingress": [...]}` (endpoint)
            continue
        if key not in all_keys:
            logger.warning(f"{key} in backend_options is not a recognized key, it will be ignored")
        if key in ("firewall", "send_prometheus_metrics", "prometheus_write"):
            logger.warning(f"{key} in backend_options is deprecated")

    # validate that we don't have aws and gcp-specific keys
    present_aws_keys = [key for key in backend_options.keys() if key in aws_keys - gcp_keys]
    present_gcp_keys = [key for key in backend_options.keys() if key in gcp_keys - aws_keys]
    if present_aws_keys and present_gcp_keys:
        raise ValueError(
            f"backend_options cannot have both AWS specific keys ({present_aws_keys}) "
            f"and GCP specific keys ({present_gcp_keys}"
        )

    # validate firewall options
    if backend_options.get("ingress"):
        if not isinstance(backend_options["ingress"], Iterable):
            raise ValueError("ingress in backend_options must be an iterable")

        firewall_keys = FirewallOptions.__optional_keys__  # type: ignore
        for fw in backend_options["ingress"]:
            for key in fw.keys():
                if key not in firewall_keys:
                    logger.warning(f"{key} ({fw[key]}) in backend_options firewall config is not a recognized key")

            for required_key in ("cidr", "ports"):
                if not fw.get(required_key):
                    raise ValueError(f"{required_key} is required for each firewall config")

            if not isinstance(fw["cidr"], str):
                raise ValueError(f"cidr ({fw['cidr']}) must be a string")
            if not isinstance(fw["ports"], list):
                raise ValueError(f"ports ({fw['ports']} must be a list")

    # validate optional bool, dict, int, str types
    all_key_types = {
        **AWSOptions.__annotations__,
        **GCPOptions.__annotations__,
    }
    keys_by_type = {
        t: [k for k, key_type in all_key_types.items() if key_type == Optional[t]] for t in (bool, dict, int, str)
    }
    for t, schema_keys in keys_by_type.items():
        for key in backend_options.keys():
            if key in schema_keys:
                val = backend_options[key]
                # all keys as optional, but if specified, they need to match desired type
                if val is not None and not isinstance(val, t):
                    raise ValueError(
                        f"{key} ({val}) in backend_options should be {t.__name__} or None, not {type(val).__name__}"
                    )


def validate_vm_typing(vm_types: Union[List[str], Tuple[str], None]):
    """Validate instance typing.

    We need to add this function because the error that our API is returning
    isn't exactly user friendly. We should raise a type error with an informative
    message instead of the dictionary that the API throws.

    """
    if not vm_types:
        return
    if not isinstance(vm_types, (list, tuple)):
        raise TypeError(
            f"Instance types must be a list or tuple, but the value '{vm_types}' is of "
            f"type: {type(vm_types)}. Please use a list of strings when specifying "
            "instance types."
        )
    for instance in vm_types:
        if not isinstance(instance, str):
            raise TypeError(
                f" Instance types must be a string, but '{instance}' is of type "
                f"{type(instance)}. Please use a string instead."
            )


def name_to_version(name: str) -> Version:
    matched = COILED_RUNTIME_REGEX.match(name)
    if matched is None:
        return Version("0.0.0")  # Should not happen
    return Version(matched.group("version").replace("-", "."))


def get_details_url(server, account, cluster_id):
    if cluster_id is None:
        return None
    else:
        return f"{server}/clusters/{cluster_id}?account={account}"


def get_grafana_url(
    cluster_details,
    account,
    cluster_id,
    grafana_server="grafana.dev-sandbox.coiledhq.com",
):
    cluster_name = cluster_details["name"]

    # for stopped clusters, only get metrics until the cluster stopped
    if cluster_details["current_state"]["state"] in ("stopped", "error"):
        end_ts = int(datetime.fromisoformat(cluster_details["current_state"]["updated"]).timestamp() * 1000)
    else:
        end_ts = "now"

    start_ts = int(datetime.fromisoformat(cluster_details["scheduler"]["created"]).timestamp() * 1000)

    base_url = f"https://{grafana_server}/d/eU1bT-nVz/cluster-metrics-prometheus?orgId=1"
    datasource = "default"  # FIXME
    # FIXME -- get account name for cluster

    full_url = (
        f"{base_url}"
        f"&datasource={datasource}"
        f"&var-env=All"
        f"&var-account={account}"
        f"&var-cluster={cluster_name}"
        f"&var-cluster-id={cluster_id}"
        f"&from={start_ts}"
        f"&to={end_ts}"
    )

    return full_url


def cluster_firewall(
    target: str,
    scheduler: int = 8786,
    dashboard: Optional[int] = 8787,
    jupyter: bool = False,
    ssh: bool = False,
    http: bool = False,
    https: bool = True,
):
    """
    Easier cluster firewall configuration when using a single CIDR.

    Examples
    --------
    To create a cluster than only accepts connections from your IP address,
    and opens port 22 for SSH access:
    >>> coiled.Cluster(
            ...,
            backend_options={**coiled.utils.cluster_firewall("me", ssh=True)}
        )

    Parameters
    ----------
    target: str
        Open cluster firewall to this range. You can either specify a CIDR,
        or use "me" to automatically get your IP address and use that, or
        "everyone" for "0.0.0.0/0".
    scheduler: int
        Port to open for access to scheduler, 8786 by default.
    dashboard: Optional[int]
        Port to open for access to dashboard, 8787 by default.
        Use `None` to not open a port for dashboard access.
    jupyter: bool
        Open port used for jupyter (8888), closed by default.
    ssh: bool
        Open port used for ssh (22), closed by default.
    http: bool
        Open port used for http (80), closed by default.
    https: bool
        Open port used for https (443), open by default.
    """
    if target == "everyone":
        cidr = "0.0.0.0/0"
    elif target == "me":
        # get public/internet routable address from which local user will be hitting scheduler
        with urllib3.PoolManager() as pool:
            my_public_ip = pool.request("GET", "https://api.ipify.org").data.decode("utf-8")

        cidr = f"{my_public_ip}/32"
    else:
        # TODO validate this this is cidr
        cidr = target

    ports = [scheduler]

    if dashboard:
        ports.append(dashboard)

    if ssh:
        ports.append(22)
    if jupyter:
        ports.append(8888)
    if http:
        ports.append(80)
    if https:
        ports.append(443)
    return {"ingress": [{"ports": ports, "cidr": cidr}]}


def is_legal_python_filename(filename: str):
    return filename.endswith(".py") and filename[:-3].isidentifier()


def recurse_importable_python_files(src: Path):
    skip_dirs = set()
    if platform.system() == "Darwin":
        skip_dirs = {
            (Path.home() / "Applications").resolve(),
            (Path.home() / "Desktop").resolve(),
            (Path.home() / "Documents").resolve(),
            (Path.home() / "Downloads").resolve(),
            (Path.home() / "Library").resolve(),
            (Path.home() / "Movies").resolve(),
            (Path.home() / "Music").resolve(),
            (Path.home() / "Pictures").resolve(),
            (Path.home() / "Public").resolve(),
        }
    dir_has_py = {}
    dot_path = Path(".")
    for root, dirs, files in os.walk(src, topdown=True):
        root_path = Path(root)
        rel_root = root_path.relative_to(src)
        has_py = False
        # Don't need .get because we always iterate through parents before
        # children, and the only case where there isn't a parent is when
        # rel_root == dot_path.
        parent_has_py = rel_root == dot_path or dir_has_py[rel_root.parent]

        for file in files:
            if is_legal_python_filename(file):
                has_py = True
                yield rel_root / file

        dir_has_py[rel_root] = has_py
        for d in dirs[:]:
            if (
                not d.isidentifier()
                or (root_path / d).resolve() in skip_dirs
                # Let's assume we can stop looking if there are no Python files
                # three levels deep.
                # For example, ./a/b/c/module.py would only be yielded if
                # there's a .py file in ./a/ or ./a/b/.
                or (not has_py and not parent_has_py and rel_root.parent != dot_path)
            ):
                # pruning inplace like this works as long as topdown=True
                dirs.remove(d)


async def validate_wheel(wheel: Path, src: str) -> Tuple[bool, str, Set[str]]:
    """
    Validate a wheel contains some python files and return the md5

    Also, warn if there are Python files in src directory that are not
    in the wheel.
    """
    hash = md5()
    has_python = False
    src_python_files = set()
    # Check if src is a URL and not a local path. This works for both Unix and
    # Windows because the drive letter in windows will get parsed as the scheme
    # and its length will be 1.
    if len(urlparse(src).scheme) > 1:
        src_path = Path(src)
        if (src_path / "src").exists():
            src_path = src_path / "src"
        for p in recurse_importable_python_files(src_path):
            parent_names = {parent.name for parent in p.parents}
            if not parent_names.intersection(IGNORE_PYTHON_DIRS):
                src_python_files.add(str(p))

    with PyZipFile(str(wheel), mode="r") as wheelzip:
        info = wheelzip.infolist()
        for file in info:
            src_python_files.discard(file.filename)

            if not has_python and file.filename != "__init__.py" and file.filename.endswith(".py"):
                has_python = True
            if "dist-info" not in file.filename:
                hash.update(str(file.CRC).encode())

    return has_python, hash.hexdigest(), src_python_files


def get_aws_identity():
    import boto3
    import botocore

    response = {}

    session = boto3.Session()
    sts = session.client("sts")

    def get_account_alias() -> str:
        try:
            iam = session.client("iam")
            r = iam.list_account_aliases()
            return r.get("AccountAliases", [])[0]
        except Exception:
            return ""

    try:
        identity = sts.get_caller_identity()

        response["account"] = identity.get("Account")
        response["who"] = identity.get("Arn").split(":")[-1]

        alias = get_account_alias()
        if alias:
            response["account_alias"] = alias
    except botocore.exceptions.NoCredentialsError:
        response["error"] = "NoCredentialsError"
    except Exception as e:
        response["error"] = str(e)

    return response


def parse_file_uri(file_uri: str) -> Path:
    p = urlparse(file_uri)
    unquoted_path = unquote(p.path)
    if os.name == "nt":
        # handling valid windows URIs of
        # file:///C:/somewhere/something.whl
        unquoted_path = unquoted_path.lstrip("/")
    return Path(unquoted_path)


def is_system_python() -> bool:
    """Determine if the current Python executable is a system Python."""
    # Check if we are in a virtualenv
    if (
        "VIRTUAL_ENV" in os.environ
        or "PYENV_VERSION" in os.environ
        or getattr(sys, "base_prefix", sys.prefix) != sys.prefix
        or getattr(sys, "real_prefix", sys.prefix) != sys.prefix
        or os.path.exists(os.path.join(sys.prefix, "conda-meta"))
    ):
        return False
    # Check installation directory for each OS
    system = platform.system()
    if system == "Linux" and sys.prefix == "/usr":
        return True
    if system == "Darwin" and (sys.prefix.startswith("/Library") or sys.prefix.startswith("/System")):
        return True
    # Default to False
    return False


NOT_INTERESTING_STACK_CODE = (
    "distributed/utils.py",
    "coiled/context.py",
    "tornado/gen.py",
    "backoff/_async.py",
    "aiohttp/client.py",
    "yarl/_url.py",
)

NOT_INTERESTING_FUNCTION = (
    "_depaginate",
    "_do_request",
    "_sync",
    "sync",
)


def is_interesting_stack_frame(filename, function):
    if any(f in filename for f in NOT_INTERESTING_STACK_CODE):
        return False
    if "coiled" in filename and (any(f in function for f in NOT_INTERESTING_FUNCTION) or function.endswith("_page")):
        return False
    return True


def truncate_traceback(exc_traceback):
    curr = exc_traceback
    frames = []
    shown = set()

    while curr:
        filename = curr.tb_frame.f_code.co_filename
        function = curr.tb_frame.f_code.co_name

        if len(frames) == 0:
            # always keep first frame
            frames.append(curr)
            shown.add((filename, function))
        elif function.startswith("_") and (filename, function[1:]) in shown:
            # ignore _foo if we've already included foo
            pass
        elif is_interesting_stack_frame(filename, function):
            frames.append(curr)
            shown.add((filename, function))

        curr = curr.tb_next

    curr = None
    for tb in reversed(frames):
        tb.tb_next = curr
        curr = tb

    return curr


def error_info_for_tracking(error: Optional[BaseException] = None) -> dict:
    loc = {}
    if error:
        if error.__traceback__:
            error_trace = "\n".join(
                [
                    line if COILED_DIR in line else "...non coiled code..."
                    for line in traceback.format_tb(error.__traceback__)
                ]
            )
        else:
            error_trace = None
        loc = {
            "error_class": error.__class__.__name__,
            "error_message": str(error),
            "error_filename": "",
            "error_line": "",
            "error_trace": error_trace,
        }
        try:
            if error.__traceback__:
                loc["error_filename"] = error.__traceback__.tb_next.tb_next.tb_frame.f_code.co_filename  # type: ignore
                loc["error_line"] = str(error.__traceback__.tb_next.tb_next.tb_frame.f_lineno)  # type: ignore
        except Exception:
            pass
    return loc


def unset_single_thread_defaults() -> dict:
    """
    Returns the env vars required to unset the default pre-spawn config that makes certain libraries
    run in single-thread mode.
    """
    env = {}
    # there are libraries that distributed by default sets to be single-threaded
    # we want to unset these default values since we're running single task on a (potentially) big machine
    for key in ("OMP_NUM_THREADS", "MKL_NUM_THREADS", "OPENBLAS_NUM_THREADS"):
        if dask.config.get(f"distributed.nanny.pre-spawn-environ.{key}", 1) == 1:
            env = {key: "", **env}
    return env


def short_random_string():
    return str(uuid.uuid4())[:8]



import asyncio
import traceback

import click
from rich import print
from rich.panel import Panel
from rich.prompt import Prompt

import coiled

from ..utils import CONTEXT_SETTINGS
from .util import setup_failure


@click.command(context_settings=CONTEXT_SETTINGS)
def setup_wizard() -> bool:
    return asyncio.run(do_setup_wizard())


async def do_setup_wizard() -> bool:
    await coiled.utils.login_if_required()

    coiled.add_interaction(
        action="cli-setup-wizard:start",
        success=True,
    )

    print(
        Panel(
            """
[bold]Welcome to Coiled![/bold]

To begin you need to connect Coiled to your cloud account.
Select one of the following options:

  1. Amazon Web Service ([green]AWS[/green])
  2. Google Cloud Platform ([green]GCP[/green])
  3. I don't have a cloud account, set me up with a free trial?
  [red]x[/red]. Exit setup

""".strip(),
            width=90,
        )
    )

    try:
        choice = Prompt.ask(
            "Choice",
            choices=["1", "2", "3", "x"],
            show_choices=False,
        )
    except KeyboardInterrupt:
        coiled.add_interaction(action="cli-setup-wizard:KeyboardInterrupt", success=False)
        return False

    coiled.add_interaction(action="cli-setup-wizard:prompt", success=True, choice=choice)

    if choice == "1":  # AWS
        from .aws import do_setup

        try:
            return do_setup(slug="coiled")
        except Exception:
            setup_failure(f"Exception raised {traceback.format_exc()}", backend="aws")
            raise
    elif choice == "2":  # GCP
        print("\nRunning [green]coiled setup gcp[/green]\n")
        from .gcp import do_setup

        try:
            return do_setup()
        except Exception:
            setup_failure(f"Exception raised {traceback.format_exc()}", backend="gcp")
            raise
    elif choice == "3":  # Other
        print(
            "\nYou can easily set up a free account with AWS or GCP.\n\n"
            "[green]AWS[/green]: [link]https://aws.amazon.com/free[/link]\n"
            "[green]GCP[/green]: [link]https://cloud.google.com/free/[/link]\n\n"
            "Alternatively, ask for a temporary guest account. "
            "Reach out at "
            "[link=mailto:hello@coiled.io?subject=Trial%20Account]hello@coiled.io[/link] "
            "with a brief note about how you'd like to use Dask and Coiled"
        )

    return False



import asyncio
import logging
import platform
import re
import shutil
import subprocess
import sys
import typing
from collections import defaultdict
from concurrent.futures import ThreadPoolExecutor
from contextlib import asynccontextmanager
from logging import getLogger
from pathlib import Path
from sys import executable
from tempfile import TemporaryDirectory
from threading import Lock
from typing import Dict, List, Optional, Set, TextIO, Tuple, Union, cast
from urllib.parse import urlparse

from dask import config
from filelock import BaseFileLock, FileLock
from packaging.utils import parse_wheel_filename
from rich.progress import Progress
from typing_extensions import Literal

from coiled.context import track_context
from coiled.scan import PackageInfo, scan_prefix
from coiled.types import ArchitectureTypesEnum, PackageLevelEnum, ResolvedPackageInfo
from coiled.utils import COILED_LOCAL_PACKAGE_PREFIX, recurse_importable_python_files, validate_wheel
from coiled.v2.core import CloudV2
from coiled.v2.widgets.util import simple_progress

logger = getLogger("coiled.package_sync")
subdir_datas = {}
PYTHON_VERSION = platform.python_version_tuple()
ANY_AVAILABLE = "ANY-AVAILABLE"


async def create_subprocess_shell(
    cmd: str,
    stdout: Union[TextIO, int, None] = None,
    stderr: Union[TextIO, int, None] = None,
) -> subprocess.CompletedProcess:
    # asyncio.create_subprocess_shell is
    # not supported on windows
    # and broken on py3.7
    loop = asyncio.get_running_loop()
    result = loop.run_in_executor(
        None, lambda: subprocess.run(cmd, shell=True, stdout=stdout, stderr=stderr, close_fds=True)
    )
    return await result


class PackageBuildError(Exception):
    pass


WHEEL_BUILD_LOCKS: Dict[str, Tuple[BaseFileLock, Lock, TemporaryDirectory]] = {}


async def default_python() -> PackageInfo:
    python_version = platform.python_version()
    return {
        "name": "python",
        "path": None,
        "source": "conda",
        "channel_url": ANY_AVAILABLE,
        "channel": ANY_AVAILABLE,
        "subdir": "linux-64",
        "conda_name": "python",
        "version": python_version,
        "wheel_target": None,
    }


# filelock is thread local
# so we have to ensure the lock is acquired/released
# on the same thread
FILE_LOCK_POOL = ThreadPoolExecutor(max_workers=1)
THREAD_LOCK_POOL = ThreadPoolExecutor(max_workers=1)


@asynccontextmanager
async def async_lock(file_lock: BaseFileLock, thread_lock: Lock):
    # Beware, there are some complicated details to this locking implementation!
    # We're trying to manage the weirdness of the file lock mostly.
    loop = asyncio.get_event_loop()
    # first acquire a thread lock
    await loop.run_in_executor(THREAD_LOCK_POOL, thread_lock.acquire)
    # acquire the file lock, we should be the only thread trying to get it
    # the threadpool is required to release it, so another thread
    # attempting to get the lock will deadlock things by preventing the
    # release!
    await loop.run_in_executor(FILE_LOCK_POOL, file_lock.acquire)
    yield
    # release the file lock first
    await loop.run_in_executor(FILE_LOCK_POOL, file_lock.release)
    # now release the thread lock, allowing another thread to proceed
    # and get the file lock
    thread_lock.release()


@track_context
async def create_wheel(pkg_name: str, version: str, src: str) -> ResolvedPackageInfo:
    # These locks are set up such that
    # Threads: Block on each other and check if another thread already built the wheel
    # Processes: Block on each other, but will not reuse a wheel created by another
    # `pip wheel` is never run on the same package at the same time
    lock_path = Path(config.PATH)
    lock_path.mkdir(parents=True, exist_ok=True)  # ensure lockfile directory exists
    package_lock, thread_lock, tmpdir = WHEEL_BUILD_LOCKS.setdefault(
        pkg_name,
        (FileLock(lock_path / ("." + pkg_name + version + ".build-lock")), Lock(), TemporaryDirectory()),
    )
    async with async_lock(package_lock, thread_lock):
        outdir = Path(tmpdir.name) / Path(pkg_name)
        if outdir.exists():
            logger.debug(f"Checking for existing wheel for {pkg_name} @ {outdir}")
            wheel_fn = next((file for file in outdir.iterdir() if file.suffix == ".whl"), None)
        else:
            wheel_fn = None
        if not wheel_fn:
            logger.debug(f"No existing wheel, creating a wheel for {pkg_name} @ {src}")
            # must use executable to avoid using some other random python
            p = await create_subprocess_shell(
                cmd=f"{executable} -m pip wheel --wheel-dir {outdir} --no-deps --use-pep517 --no-cache-dir {src}",
                stderr=subprocess.STDOUT,
                stdout=subprocess.PIPE,
            )
            if p.returncode:
                print(
                    f"---Wheel Build Log for {pkg_name}---\n" + p.stdout.decode(encoding=sys.stdout.encoding or "utf-8")
                )
                return {
                    "name": pkg_name,
                    "source": "pip",
                    "channel": None,
                    "conda_name": None,
                    "client_version": version,
                    "specifier": "",
                    "include": False,
                    "error": (
                        "Failed to build a wheel for the"
                        " package, will not be included in environment, check stdout for the build log"
                    ),
                    "note": None,
                    "sdist": None,
                    "md5": None,
                }
            wheel_fn = next(file for file in outdir.iterdir() if file.suffix == ".whl")
        logger.debug(f"Using wheel @ {wheel_fn}")
        _, build_version, _, _ = parse_wheel_filename(str(wheel_fn.name))
        has_python, md5, missing_py_files = await validate_wheel(wheel_fn, src)
    return {
        "name": pkg_name,
        "source": "pip",
        "channel": None,
        "conda_name": None,
        "client_version": str(build_version),
        "specifier": "",
        "include": True,
        "error": None if has_python else "Built wheel contains no python files!",
        "note": (
            f"Wheel built from {src}"
            + (f" is missing {', '.join(sorted(missing_py_files)[:10])}" if missing_py_files else "")
        ),
        "sdist": wheel_fn.open("rb"),
        "md5": md5,
    }


@track_context
async def create_wheel_from_egg(pkg_name: str, version: str, src: str) -> ResolvedPackageInfo:
    tmpdir = TemporaryDirectory()
    outdir = Path(tmpdir.name) / Path(pkg_name)
    outdir.mkdir(parents=True)
    logger.debug(f"Attempting to create a wheel for {pkg_name} in directory {src}")
    # must use executable to avoid using some other random python
    p = await create_subprocess_shell(
        cmd=f"{executable} -m wheel convert --dest-dir {outdir} {src}",
        stderr=subprocess.STDOUT,
        stdout=subprocess.PIPE,
    )
    if p.returncode:
        print(
            f"---Egg to wheel conversion Log for {pkg_name}---\n"
            + p.stdout.decode(encoding=sys.stdout.encoding or "utf-8")
        )
        return {
            "name": pkg_name,
            "source": "pip",
            "channel": None,
            "conda_name": None,
            "client_version": version,
            "specifier": "",
            "include": False,
            "error": (
                "Failed to convert the package egg to a wheel"
                ", will not be included in environment, check stdout for egg conversion log"
            ),
            "note": None,
            "sdist": None,
            "md5": None,
        }
    wheel_fn = next(file for file in outdir.iterdir() if file.suffix == ".whl")
    has_python, md5, missing_py_files = await validate_wheel(Path(wheel_fn), tmpdir.name)
    return {
        "name": pkg_name,
        "source": "pip",
        "channel": None,
        "conda_name": None,
        "client_version": version,
        "specifier": "",
        "include": True,
        "error": None if has_python else "Built wheel has no python files!",
        "note": (
            "Wheel built from local egg"
            + (f" is missing {', '.join(sorted(missing_py_files)[:10])}" if missing_py_files else "")
        ),
        "sdist": wheel_fn.open("rb"),
        "md5": md5,
    }


@track_context
async def create_wheel_from_src_dir(pkg_name: str, version: str, src: str) -> ResolvedPackageInfo:
    # These locks are set up such that
    # Threads: Block on each other and check if another thread already built the tarball
    # Processes: Block on each other, but will not reuse a tarball created by another
    md5 = None
    lock_path = Path(config.PATH)
    lock_path.mkdir(parents=True, exist_ok=True)  # ensure lockfile directory exists
    package_lock, thread_lock, tmpdir = WHEEL_BUILD_LOCKS.setdefault(
        pkg_name,
        (FileLock(lock_path / (f".{pkg_name}{version}.build-lock")), Lock(), TemporaryDirectory()),
    )
    async with async_lock(package_lock, thread_lock):
        outdir = Path(tmpdir.name) / Path(pkg_name)
        if outdir.exists():
            logger.debug(f"Checking for existing source archive for {pkg_name} @ {outdir}")
            wheel_fn = next((file for file in outdir.iterdir() if file.suffix == ".whl"), None)
        else:
            wheel_fn = None
        if not wheel_fn:
            logger.debug(f"No existing source archive, creating an archive for {pkg_name} @ {src}")
            try:
                unpacked_dir = outdir / f"{pkg_name}-{version}"
                # Create fake metadata to make wheel work
                dist_info_dir = unpacked_dir / f"{unpacked_dir.name}.dist-info"
                dist_info_dir.mkdir(parents=True)
                with open(dist_info_dir / "METADATA", "w") as f:
                    f.write(f"Metadata-Version: 2.1\nName: {pkg_name}\nVersion: {version}\n")
                with open(dist_info_dir / "WHEEL", "w") as f:
                    f.write("Wheel-Version: 1.0\nGenerator: coiled\nRoot-Is-Purelib: true\nTag: py3-none-any\n")
                src_path = Path(src)
                for file in recurse_importable_python_files(src_path):
                    if str(file) in ("__init__.py", "__main__.py"):
                        continue
                    dest = unpacked_dir / file
                    dest.parent.mkdir(parents=True, exist_ok=True)
                    shutil.copy(src_path / file, dest)
                p = await create_subprocess_shell(
                    cmd=f"{executable} -m wheel pack --dest-dir {outdir} {unpacked_dir}",
                    stderr=subprocess.STDOUT,
                    stdout=subprocess.PIPE,
                )
                if p.returncode:
                    print(
                        f"---wheel packing log for {src}---\n"
                        + p.stdout.decode(encoding=sys.stdout.encoding or "utf-8")
                    )
                    return {
                        "name": pkg_name,
                        "source": "pip",
                        "channel": None,
                        "conda_name": None,
                        "client_version": version,
                        "specifier": "",
                        "include": False,
                        "error": (
                            "Failed to build a package of your local python files. Please check stdout for details"
                        ),
                        "note": None,
                        "sdist": None,
                        "md5": None,
                    }
            except IOError as e:
                return {
                    "name": pkg_name,
                    "source": "pip",
                    "channel": None,
                    "conda_name": None,
                    "client_version": version,
                    "specifier": "",
                    "include": False,
                    "error": f"Failed to build a package of your local python files. Exception: {e}",
                    "note": None,
                    "sdist": None,
                    "md5": None,
                }
            wheel_fn = next(file for file in outdir.iterdir() if file.suffix == ".whl")
        logger.debug(f"Using wheel @ {wheel_fn}")
        _, build_version, _, _ = parse_wheel_filename(str(wheel_fn.name))
        has_python, md5, missing_py_files = await validate_wheel(wheel_fn, src)
    return {
        "name": pkg_name,
        "source": "pip",
        "channel": None,
        "conda_name": None,
        "client_version": str(build_version),
        "specifier": "",
        "include": True,
        "error": None if has_python else "Built wheel does not contain all python files!",
        "note": (
            f"Source wheel built from {src}"
            + (f" is missing {', '.join(sorted(missing_py_files)[:10])}" if missing_py_files else "")
        ),
        "sdist": wheel_fn.open("rb"),
        "md5": md5,
    }


@track_context
async def approximate_packages(
    cloud: CloudV2,
    packages: List[PackageInfo],
    priorities: Dict[Tuple[str, Literal["conda", "pip"]], PackageLevelEnum],
    progress: Optional[Progress] = None,
    strict: bool = False,
    architecture: ArchitectureTypesEnum = ArchitectureTypesEnum.X86_64,
) -> typing.List[ResolvedPackageInfo]:
    user_conda_installed_python = next((p for p in packages if p["name"] == "python"), None)
    user_conda_installed_pip = next(
        (i for i, p in enumerate(packages) if p["name"] == "pip" and p["source"] == "conda"),
        None,
    )
    if not user_conda_installed_pip:
        # This means pip was installed by pip, or the system
        # package manager
        # Insert a conda version of pip to be installed first, it will
        # then be used to install the users version of pip
        pip = next(
            (p for p in packages if p["name"] == "pip" and p["source"] == "pip"),
            None,
        )
        if not pip:
            # insert a modern version and hope it does not introduce conflicts
            packages.append(
                {
                    "name": "pip",
                    "path": None,
                    "source": "conda",
                    "channel_url": "https://conda.anaconda.org/conda-forge/",
                    "channel": "conda-forge",
                    "subdir": "noarch",
                    "conda_name": "pip",
                    "version": "22.3.1",
                    "wheel_target": None,
                }
            )
        else:
            # insert the users pip version and hope it exists on conda-forge
            packages.append(
                {
                    "name": "pip",
                    "path": None,
                    "source": "conda",
                    "channel_url": "https://conda.anaconda.org/conda-forge/",
                    "channel": "conda-forge",
                    "subdir": "noarch",
                    "conda_name": "pip",
                    "version": pip["version"],
                    "wheel_target": None,
                }
            )
    coiled_selected_python = None
    if not user_conda_installed_python:
        # insert a special python package
        # that the backend will pick a channel for
        coiled_selected_python = await default_python()
        packages.append(coiled_selected_python)

    local_packages: List[PackageInfo] = [pkg for pkg in packages if pkg["name"].startswith(COILED_LOCAL_PACKAGE_PREFIX)]
    packages = [pkg for pkg in packages if not pkg["name"].startswith(COILED_LOCAL_PACKAGE_PREFIX)]

    with simple_progress("Validating environment", progress=progress):
        results = await cloud._approximate_packages(
            packages=[
                {
                    "name": pkg["name"],
                    "priority_override": (
                        PackageLevelEnum.CRITICAL
                        if (
                            strict
                            or (
                                pkg["wheel_target"]
                                # Ignore should override wheel_target (see #2640)
                                and not priorities.get((pkg["name"], pkg["source"])) == PackageLevelEnum.IGNORE
                            )
                        )
                        else priorities.get(
                            ((cast(str, pkg["conda_name"]) if pkg["source"] == "conda" else pkg["name"]), pkg["source"])
                        )
                    ),
                    "python_major_version": PYTHON_VERSION[0],
                    "python_minor_version": PYTHON_VERSION[1],
                    "python_patch_version": PYTHON_VERSION[2],
                    "source": pkg["source"],
                    "channel_url": pkg["channel_url"],
                    "channel": pkg["channel"],
                    "subdir": pkg["subdir"],
                    "conda_name": pkg["conda_name"],
                    "version": pkg["version"],
                    "wheel_target": pkg["wheel_target"],
                }
                for pkg in packages
            ],
            architecture=architecture,
        )
    result_map = {(r["name"], r["conda_name"]): r for r in results}
    finalized_packages: typing.List[ResolvedPackageInfo] = []

    if not user_conda_installed_python and coiled_selected_python:
        # user has no python version installed by conda
        # we should have a result of asking the backend to
        # pick conda channel that has the users python version
        python_result = result_map[("python", "python")]
        if result_map[("python", "python")]["error"]:
            finalized_packages.append(
                {
                    "name": "python",
                    "source": "conda",
                    "channel": None,
                    "conda_name": "python",
                    "client_version": coiled_selected_python["version"],
                    "specifier": python_result["specifier"] or "",
                    "include": python_result["include"],
                    "note": None,
                    "error": python_result["error"],
                    "sdist": None,
                    "md5": None,
                }
            )
        else:
            note = python_result["note"]
            if not note:
                raise ValueError("Expected a note from the backend")
            channel_url, channel = note.split(",")
            finalized_packages.append(
                {
                    "name": "python",
                    "source": "conda",
                    "channel": channel,
                    "conda_name": "python",
                    "client_version": coiled_selected_python["version"],
                    "specifier": python_result["specifier"] or "",
                    "include": python_result["include"],
                    "note": None,
                    "error": python_result["error"],
                    "sdist": None,
                    "md5": None,
                }
            )
        # we can pull our special python package out of the list
        # now
        packages.remove(coiled_selected_python)
    for pkg in local_packages:
        if pkg["wheel_target"]:
            with simple_progress(f'Creating wheel for {pkg["wheel_target"]}', progress=progress):
                finalized_packages.append(
                    await create_wheel_from_src_dir(
                        pkg_name=pkg["name"],
                        version=pkg["version"],
                        src=pkg["wheel_target"],
                    )
                )
    for pkg in packages:
        package_result = result_map[(pkg["name"], pkg["conda_name"])]
        if pkg["wheel_target"] and package_result["include"]:
            p = urlparse(pkg["wheel_target"])
            if len(p.scheme) <= 1 and not Path(pkg["wheel_target"]).exists():
                # lack of scheme (or it being a 1-char drive letter) means a local file
                # sometimes the wheel target taken from
                # direct_url.json references a file that does not exist
                # skip over trying to sync that and treat it like a normal package
                # this can happen in conda/system packages
                # where the package metadata was generated on a build system and not locally
                finalized_packages.append(
                    {
                        "name": pkg["name"],
                        "source": pkg["source"],
                        "channel": pkg["channel"],
                        "conda_name": pkg["conda_name"],
                        "client_version": pkg["version"],
                        "specifier": package_result["specifier"] or "",
                        "include": package_result["include"],
                        "note": package_result["note"],
                        "error": package_result["error"],
                        "sdist": None,
                        "md5": None,
                    }
                )
            elif pkg["wheel_target"].endswith(".egg"):
                with simple_progress(f'Creating wheel from egg for {pkg["name"]}', progress=progress):
                    finalized_packages.append(
                        await create_wheel_from_egg(
                            pkg_name=pkg["name"],
                            version=pkg["version"],
                            src=pkg["wheel_target"],
                        )
                    )
            else:
                with simple_progress(f'Creating wheel for {pkg["name"]}', progress=progress):
                    finalized_packages.append(
                        await create_wheel(
                            pkg_name=pkg["name"],
                            version=pkg["version"],
                            src=pkg["wheel_target"],
                        )
                    )
        else:
            finalized_packages.append(
                {
                    "name": pkg["name"],
                    "source": pkg["source"],
                    "channel": pkg["channel"],
                    "conda_name": pkg["conda_name"],
                    "client_version": pkg["version"],
                    "specifier": package_result["specifier"] or "",
                    "include": package_result["include"],
                    "note": package_result["note"],
                    "error": package_result["error"],
                    "sdist": None,
                    "md5": None,
                }
            )
    return finalized_packages


pip_bad_req_regex = (
    r"(?P<package>.+) (?P<version>.+) has requirement "
    r"(?P<requirement>.+), but you have (?P<requirement2>.+) (?P<reqversion>.+)."
)


@track_context
async def check_pip_happy(progress: Optional[Progress] = None) -> Dict[str, List[str]]:
    with simple_progress("Running pip check", progress=progress):
        proc = await create_subprocess_shell(
            f"{executable} -m pip check", stdout=subprocess.PIPE, stderr=subprocess.PIPE
        )
        faulty_packages = defaultdict(list)
        if proc.returncode:
            output = proc.stdout.decode(encoding=sys.stdout.encoding or "utf-8")
            bad_reqs = re.finditer(pip_bad_req_regex, output)
            for bad_req in bad_reqs:
                groups = bad_req.groupdict()
                span = bad_req.span()
                warning = output[span[0] : span[1]]
                logger.warning(warning)
                faulty_packages[groups["package"]].append(warning)
        return faulty_packages


@track_context
async def create_environment_approximation(
    cloud: CloudV2,
    priorities: Dict[Tuple[str, Literal["conda", "pip"]], PackageLevelEnum],
    only: Optional[Set[str]] = None,
    strict: bool = False,
    progress: Optional[Progress] = None,
    architecture: ArchitectureTypesEnum = ArchitectureTypesEnum.X86_64,
) -> typing.List[ResolvedPackageInfo]:
    packages = await scan_prefix(progress=progress)
    bad_packages = await check_pip_happy(progress)
    if bad_packages:
        packages = [pkg for pkg in packages if pkg["name"] not in bad_packages]
    if only:
        packages = filter(lambda pkg: pkg["name"] in only, packages)
    # TODO: private conda channels
    result = await approximate_packages(
        cloud=cloud,
        packages=[pkg for pkg in packages],
        priorities=priorities,
        strict=strict,
        progress=progress,
        architecture=architecture,
    )
    for bad_package, error_list in bad_packages.items():
        errors = "\n".join(error_list)
        result.append(
            {
                "name": bad_package,
                "error": f"Pip check had the following issues that need resolving: \n{errors}",
                "source": "pip",
                "channel": None,
                "client_version": "",
                "conda_name": None,
                "include": False,
                "md5": None,
                "note": None,
                "sdist": None,
                "specifier": "",
            }
        )
    return result


if __name__ == "__main__":
    from logging import basicConfig

    basicConfig(level=logging.INFO)

    from rich.console import Console
    from rich.table import Table

    async def run():
        async with CloudV2(asynchronous=True) as cloud:
            return await create_environment_approximation(
                cloud=cloud,
                priorities={
                    ("dask", "conda"): PackageLevelEnum.CRITICAL,
                    ("twisted", "conda"): PackageLevelEnum.IGNORE,
                    ("graphviz", "conda"): PackageLevelEnum.LOOSE,
                    ("icu", "conda"): PackageLevelEnum.LOOSE,
                },
            )

    result = asyncio.run(run())

    table = Table(title="Packages")
    keys = ("name", "source", "include", "client_version", "specifier", "error", "note")

    for key in keys:
        table.add_column(key)

    for pkg in result:
        row_values = [str(pkg.get(key, "")) for key in keys]
        table.add_row(*row_values)
    console = Console()
    console.print(table)
    console.print(table)



import enum
import logging

import dask.config
from dask.distributed import Client
from distributed.deploy.adaptive import Adaptive

from .utils import COILED_LOGGER_NAME

logger = logging.getLogger(COILED_LOGGER_NAME)


@enum.unique
class CredentialsPreferred(enum.Enum):
    LOCAL = "local"
    # USER = 'user'
    ACCOUNT = "account"  # doesn't work, should be fixed or deprecated/removed someday
    NONE = None


class CoiledAdaptive(Adaptive):
    def __init__(self, cluster, **kwargs):
        super().__init__(cluster, **kwargs)
        target_duration = kwargs.get("target_duration")
        if target_duration:
            with Client(self.cluster) as client:
                client.run_on_scheduler(
                    lambda: dask.config.set({"distributed.adaptive.target-duration": target_duration})
                )

    def config_as_string(self):
        return (
            f"Config is maximum: {self.maximum}, minimum: {self.minimum}, wait_count: {self.wait_count}, "
            f"interval: {self.interval}, target_duration: {self.target_duration}."
        )

    async def scale_up(self, n):
        logger.info(f"Adaptive scaling up to {n} workers.")
        if self.cluster:
            await self.cluster.scale_up(
                n,
                reason=f"Adaptive scaling up to {n} workers. "
                f"Previous intended size was {len(self.cluster._plan)}. "
                f"{self.config_as_string()}",
            )

    async def scale_down(self, workers):
        if not workers:
            return
        logger.info(f"Adaptive is removing {len(workers)} workers: {workers}.")
        if self.cluster:
            await self.cluster.scale_down(
                workers,
                reason=f"Adaptive removing {len(workers)} workers. "
                f"Previous intended size was {len(self.cluster._plan)}. "
                f"{self.config_as_string()}",
            )

    async def workers_to_close(self, target: int):
        if self.cluster and hasattr(self.cluster, "workers_to_close"):
            # use our logic for determining which workers to close
            # e.g., if there's an extra worker on scheduler VM, don't close that
            return await self.cluster.workers_to_close(target=target)

        # just in case, fall back to base class method
        return await super().workers_to_close(target=target)

    async def recommendations(self, target: int) -> dict:
        if self.cluster and hasattr(self.cluster, "_set_plan_requested"):
            # update current view of workers if running worker on scheduler
            # since we initially include fake worker in the set but this messes things up if we
            # don't remove later when actual worker has appeared.
            if "extra-worker-on-scheduler" in self.cluster.plan:
                await self.cluster._set_plan_requested()
        return await super().recommendations(target=target)



from __future__ import annotations

import datetime
import io
import json
import os
import shlex
import subprocess
import sys
import time
from typing import Optional, Sequence

import click
import dask.config
from dask.base import tokenize
from dask.distributed import Client, get_worker
from dask.utils import parse_timedelta
from packaging.version import Version
from rich import print

import coiled
from coiled.compatibility import DISTRIBUTED_VERSION
from coiled.spans import span
from coiled.utils import error_info_for_tracking, unset_single_thread_defaults

from ..v2.widgets.rich import LightRichClusterWidget
from .cluster.better_logs import better_logs
from .utils import CONTEXT_SETTINGS

MINIMUM_DISTRIBUTED_VERSION = Version("2023.6.0")


class KeepaliveSession:
    def __init__(self, cluster, prefix=""):
        self.cluster = cluster
        rand_uuid = coiled.utils.short_random_string()
        self.session_id = f"{prefix}-{rand_uuid}" if prefix else rand_uuid

    def __enter__(self):
        # keepalive session lets us use keepalive without dask client
        self.cluster._call_scheduler_comm("coiled_add_keepalive_session", name=self.session_id)

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.cluster._call_scheduler_comm("coiled_end_keepalive_session", name=self.session_id)


def get_ssh_connection(cloud, cluster_id):
    try:
        import fabric
        from paramiko.ed25519key import Ed25519Key
    except ImportError:
        # shouldn't get here but just in case
        raise RuntimeError("--use-ssh requires fabric")

    ssh_info = cloud.get_ssh_key(cluster_id=cluster_id)

    with io.StringIO(ssh_info["private_key"]) as f:
        pkey = Ed25519Key(file_obj=f)

    connection = fabric.connection.Connection(
        ssh_info["scheduler_public_address"], user="ubuntu", connect_kwargs={"pkey": pkey}
    )

    return connection


def write_via_ssh(connection, content, path, mode=None):
    with io.StringIO(content) as f:
        connection.put(f, path)
    if mode:
        connection.sftp().chmod(path, mode)


def run_via_ssh(cloud, cluster, info, command, file, env, interactive: bool, detach: bool):
    try:
        import invoke
    except ImportError:
        # shouldn't get here but just in case
        raise RuntimeError("--use-ssh requires fabric")

    connection = get_ssh_connection(cloud, cluster.cluster_id)
    results = None

    original_command = " ".join(command)
    callstack = [{"code": original_command, "relative_line": 0}]
    # Extract and upload files from `command`
    command = shlex.split(original_command)
    info["files-implicit"] = []
    for idx, i in enumerate(command):
        if os.path.exists(i) and os.path.isfile(i):
            # flatten paths, everything will be /scratch/foo
            # without any subdirectories
            connection.put(i, "/scratch/")
            info["files-implicit"].append(i)
            # adjust command to reference path on VM
            command[idx] = f"/scratch/{os.path.basename(i)}"

            # include code files in the "callstack"
            if i.endswith((".py", ".sh")):
                try:
                    with open(i) as f:
                        contents = f.read()
                    callstack.append({"code": contents, "filename": i})
                except Exception:
                    pass

    # Upload user-specified files too
    info["files-explicit"] = file
    for f in file:
        connection.put(f, "/scratch/")

    command_string = " ".join(command)

    container_name = "tmp-dask-1"

    entrypoint = get_entrypoint(connection, container_name)
    if "cloud-env" in entrypoint:
        # hack for cloud-env containers (e.g., package sync) currently required to avoid re-downloading senv
        # once container is suitably changed, we should just be able to use entrypoint like any other container
        entrypoint = "micromamba run -p /opt/coiled/env"

    docker_opts = "-it" if interactive else ""
    docker_opts = f"{docker_opts} --detach" if detach else docker_opts

    env_opts = " ".join(f"--env {key}={val}" for key, val in env.items()) if env else ""

    if not interactive:
        command_id = f"{datetime.datetime.now().isoformat()}_{coiled.utils.short_random_string()}"

        command_dir = "/scratch/run"
        inner_command_path = f"{command_dir}/{command_id}.sh"
        tee_command = f"{command_dir}/{command_id}-tee.sh"

        connection.run(f"mkdir -p {command_dir}")

        # write the user's command as a shell file
        write_via_ssh(connection, command_string, inner_command_path, mode=0o755)
        # wrapper file that calls user command, sending output to pseudo-tty and stdout (which goes into logs)
        write_via_ssh(connection, f"{inner_command_path} | tee -a /proc/1/fd/1", tee_command, mode=0o755)
        command_string = tee_command

    container_command = (
        f"docker exec --workdir /scratch {env_opts} {docker_opts} {container_name} {entrypoint} {command_string}"
    )
    try:
        with KeepaliveSession(cluster=cluster, prefix="ssh"), span(cluster, callstack=callstack):
            results = connection.run(container_command, hide=not interactive, pty=interactive)
    except invoke.exceptions.UnexpectedExit as e:
        results = str(e)
    except Exception as e:
        raise e

    return connection, container_command, results


def get_entrypoint(connection, container_name) -> str:
    entrypoint = ""
    x = connection.run(f"docker inspect -f '{{{{json .Config.Entrypoint}}}}' {container_name}", hide=True)
    if x.stdout:
        entrypoint = json.loads(x.stdout)

        if entrypoint:
            # hack so we don't use `tini` for each run, remove everything between `tini` and `--`
            # e.g., "tini -g -- /usr/bin/prepare.sh" -> "/usr/bin/prepare.sh"
            if "tini" in entrypoint[0]:
                if "--" in entrypoint:
                    after_tini = entrypoint.index("--")
                    entrypoint = entrypoint[after_tini + 1 :]

            entrypoint = " ".join(entrypoint) if entrypoint else ""
    return entrypoint or ""


@click.command(context_settings=CONTEXT_SETTINGS)
@click.option(
    "--name",
    default=None,
    help="Run name. If not given, defaults to a unique name.",
)
@click.option(
    "--account",
    default=None,
    help="Coiled account (uses default account if not specified)",
)
@click.option(
    "--software",
    default=None,
    help=(
        "Software environment name to use. If neither software nor container is specified, "
        "all the currently-installed Python packages are replicated on the VM using package sync."
    ),
)
@click.option(
    "--container",
    default=None,
    help=(
        "Container image to use. If neither software nor container is specified, "
        "all the currently-installed Python packages are replicated on the VM using package sync."
    ),
)
@click.option(
    "--vm-type",
    default=[],
    multiple=True,
    help="VM type to use. Specify multiple times to provide multiple options.",
)
@click.option(
    "--gpu",
    default=False,
    is_flag=True,
    help="Have a GPU available.",
)
@click.option(
    "--region",
    default=None,
    help="The cloud provider region in which to run the notebook.",
)
@click.option(
    "--disk-size",
    default=None,
    help="Use larger-than-default disk on VM, specified in GiB.",
)
@click.option(
    "--keepalive",
    default=None,
    help=(
        "Keep your VM running for the specified time, even after your command completes. "
        "In seconds (`--keepalive 60`) unless you specify units (`--keepalive 3m` for 3 minutes)."
        "Default to shutdown immediately after the command finishes."
    ),
)
@click.option(
    "--file",
    "-f",
    default=[],
    multiple=True,
    help="Local files required to run command.",
)
@click.option(
    "--env",
    "-e",
    default=[],
    multiple=True,
    help=(
        "Environment variables securely transmitted to run command environment. "
        "Format is `KEY=val`, multiple vars can be set with separate `--env` for each."
    ),
)
@click.option(
    "--use-ssh",
    default=False,
    is_flag=True,
    hidden=True,
)
@click.option(
    "--interactive",
    default=False,
    is_flag=True,
    hidden=True,
)
@click.option(
    "--detach",
    default=False,
    is_flag=True,
    hidden=True,
)
@click.argument("command", nargs=-1)
def run(
    name: str,
    account: Optional[str],
    software: Optional[str],
    container: Optional[str],
    vm_type: Sequence[str],
    gpu: bool,
    region: Optional[str],
    disk_size: Optional[int],
    keepalive,
    file,
    env,
    use_ssh: bool,
    interactive: bool,
    detach: bool,
    command,
):
    """
    Run a command on the cloud.
    """
    if DISTRIBUTED_VERSION < MINIMUM_DISTRIBUTED_VERSION:
        # Relies on scheduler having scratch space, which was added in https://github.com/dask/distributed/pull/7802
        raise RuntimeError(
            f"`coiled run` requires distributed>={MINIMUM_DISTRIBUTED_VERSION} "
            f"(distributed={DISTRIBUTED_VERSION} is installed)"
        )

    if use_ssh:
        # while --use-ssh is hidden option we won't add fabric to client deps
        try:
            import fabric  # noqa: F401
        except ImportError:
            print(
                "[red][bold]Error:[/bold][/red] to use [green]--use-ssh[/green] "
                "you need to install [green]fabric[/green]"
            )
            return

    env_dict = dict(env_var.split("=", maxsplit=1) for env_var in env)

    dask_env = unset_single_thread_defaults()
    if container and "rapidsai" in container:
        dask_env = {"DISABLE_JUPYTER": "true", **dask_env}  # needed for "stable" RAPIDS image

    if not command:
        raise ValueError("command must be specified")

    keepalive = parse_timedelta(keepalive)
    shutdown_on_close = keepalive is None

    info = {"command": command, "keepalive": keepalive}
    success = True
    exception = None

    # if user tries `coiled run foo.py` they probably want to run `python foo.py` rather than `foo.py`
    if len(command) == 1 and command[0].endswith(".py"):
        command = ("python", command[0])

    connection = None
    container_command = None
    results = None

    try:
        with LightRichClusterWidget(
            account=account, title=f"Running [bold]{shlex.join(command)}[/bold]"
        ) as widget, coiled.Cloud(account=account) as cloud:
            account = account or cloud.default_account
            widget.update(
                server=cloud.server,
                cluster_details=None,
                logs=None,
                account=account,
            )
            info["account"] = account
            cluster_kwargs = dict(
                account=account,
                n_workers=0,
                software=software,
                container=container,
                scheduler_options={"idle_timeout": "24 hours"},
                scheduler_vm_types=list(vm_type) if vm_type else None,
                worker_vm_types=list(vm_type) if vm_type else None,
                allow_ssh=True,
                extra_worker_on_scheduler=True,
                environ=dask_env,
                scheduler_gpu=gpu,
                region=region,
                shutdown_on_close=shutdown_on_close,
                tags={"coiled-cluster-type": "run/cli"},
                scheduler_disk_size=disk_size,
                worker_disk_size=disk_size,
            )
            token = tokenize(sys.executable, **cluster_kwargs)
            name = name or f"run-{token[:8]}"

            dask.config.set({"coiled.use_aws_creds_endpoint": True})
            with coiled.Cluster(name=name, cloud=cloud, custom_widget=widget, **cluster_kwargs) as cluster:
                info["cluster_id"] = cluster.cluster_id
                if not use_ssh:
                    with Client(cluster) as client:
                        if not shutdown_on_close:
                            cluster.set_keepalive(keepalive=keepalive)
                        if env_dict:
                            cluster.send_private_envs(env_dict)

                        original_command = " ".join(command)
                        callstack = [{"code": original_command, "relative_line": 0}]
                        # Extract and upload files from `command`
                        command = shlex.split(original_command)
                        info["files-implicit"] = []
                        for idx, i in enumerate(command):
                            if os.path.exists(i) and os.path.isfile(i):
                                client.upload_file(i, load=False)
                                info["files-implicit"].append(i)
                                command[idx] = os.path.basename(i)

                                # include code files in the "callstack"
                                if i.endswith((".py", ".sh")):
                                    try:
                                        with open(i) as f:
                                            contents = f.read()
                                        callstack.append({"code": contents, "filename": i})
                                    except Exception:
                                        pass

                        # Upload user-specified files too
                        info["files-explicit"] = file
                        for f in file:
                            client.upload_file(f, load=False)

                        def run_command(command):
                            subprocess.run(command, cwd=get_worker().local_directory)

                        info["command-parsed"] = command

                        # keep track of time before we start executing user code
                        start_ns = time.time_ns()

                        # TODO execute command in non-blocking way so we can also tail logs while it runs
                        client.submit(run_command, command=["echo", "@start-coiled-run"]).result()
                        with span(cluster, callstack=callstack):
                            client.submit(run_command, command=command).result()
                        client.submit(run_command, command=["echo", "@stop-coiled-run"]).result()
                else:
                    if not shutdown_on_close:
                        cluster.set_keepalive(keepalive=keepalive)

                    if interactive:
                        widget.stop()

                    start_ns = time.time_ns()
                    connection, container_command, results = run_via_ssh(
                        cloud=cloud,
                        cluster=cluster,
                        info=info,
                        command=command,
                        file=file,
                        interactive=interactive,
                        detach=detach,
                        env=env_dict,
                    )

        if cluster.cluster_id and not interactive:
            if detach:
                print(f"Running [green]{shlex.join(command)}[/green] in background.")
            else:
                print()

                print("Output")
                print("-" * len("Output"))
                print()

            if not use_ssh:
                better_logs(
                    cluster_id=cluster.cluster_id,
                    instance_labels_dict={},
                    show_label=False,
                    show_all_instances=True,
                    show_timestamp=False,  # only show the messages
                    tail=True,  # use tail so that we keep polling logs until stop sentinel to shows up
                    tail_max_times=10,  # only try 10 times (at 3s interval), logs should be ingested by then
                    since=str(int(start_ns // 1e6)),  # it wants timestamp in ms as a string
                    start_sentinel="@start-coiled-run",
                    stop_sentinel="@stop-coiled-run",
                )
            else:
                if results:
                    if isinstance(results, str):
                        print(results)
                    else:
                        print(results.stdout)
                        if results.stderr:
                            print(f"[red]{results.stderr}")

            print()

    except Exception as e:
        success = False
        exception = e
        raise e
    finally:
        coiled.add_interaction(
            "coiled-run-cli",
            success=success,
            **info,
            **error_info_for_tracking(exception),
        )



import re
from enum import Enum
from logging import getLogger
from pathlib import Path
from typing import BinaryIO, Dict, List, Optional, Union

from typing_extensions import Literal, TypedDict

logger = getLogger("coiled.package_sync")


event_type_list = Literal[
    "add_role_to_profile",
    "attach_gateway_to_router",
    "attach_subnet_to_router",
    "create_vm",
    "create_machine_image",
    "create_scheduler" "create_worker",
    "delete_machine_image",
    "create_fw_rule",
    "create_fw",
    "create_network_cidr",
    "create_subnet",
    "create_network",
    "create_log_sink",
    "create_router",
    "create_iam_role",
    "create_log_bucket",
    "create_storage_bucket",
    "create_instance_profile",
    "check_log_sink_exists",
    "check_or_attach_cloudwatch_policy",
    "delete_vm",
    "delete_route",
    "get_firewall",
    "get_network",
    "get_subnet",
    "get_policy_arn",
    "get_log_group",
    "gcp_instance_create",
    "net_gateways_get_or_create",
    "scale",
]


class CondaPlaceHolder(dict):
    pass


class PackageInfo(TypedDict):
    name: str
    path: Optional[Path]
    source: Literal["pip", "conda"]
    channel_url: Optional[str]
    channel: Optional[str]
    subdir: Optional[str]
    conda_name: Optional[str]
    version: str
    wheel_target: Optional[str]


class PackageSchema(TypedDict):
    name: str
    source: Literal["pip", "conda"]
    channel: Optional[str]
    conda_name: Optional[str]
    client_version: Optional[str]
    specifier: str
    include: bool
    file: Optional[int]


class ResolvedPackageInfo(TypedDict):
    name: str
    source: Literal["pip", "conda"]
    channel: Optional[str]
    conda_name: Optional[str]
    client_version: Optional[str]
    specifier: str
    include: bool
    note: Optional[str]
    error: Optional[str]
    sdist: Optional[BinaryIO]
    md5: Optional[str]


class PackageLevelEnum(int, Enum):
    """
    Package mismatch severity level
    Using a high int so we have room to add extra levels as needed

    Ordering is allow comparison like

    if somelevel >= PackageLevelEnum.STRICT_MATCH:
        <some logic for high or critical levels>
    """

    CRITICAL = 100
    STRICT_MATCH = 75
    WARN = 50
    NONE = 0
    LOOSE = -1
    IGNORE = -2


class ApproximatePackageRequest(TypedDict):
    name: str
    priority_override: Optional[PackageLevelEnum]
    python_major_version: str
    python_minor_version: str
    python_patch_version: str
    source: Literal["pip", "conda"]
    channel_url: Optional[str]
    channel: Optional[str]
    subdir: Optional[str]
    conda_name: Optional[str]
    version: str
    wheel_target: Optional[str]


class ApproximatePackageResult(TypedDict):
    name: str
    conda_name: Optional[str]
    specifier: Optional[str]
    include: bool
    note: Optional[str]
    error: Optional[str]


class PiplessCondaEnvSchema(TypedDict, total=False):
    name: Optional[str]
    channels: List[str]
    dependencies: List[str]


class CondaEnvSchema(TypedDict, total=False):
    name: Optional[str]
    channels: List[str]
    dependencies: List[Union[str, Dict[str, List[str]]]]


class SoftwareEnvSpec(TypedDict):
    packages: List[PackageSchema]
    raw_pip: Optional[List[str]]
    raw_conda: Optional[CondaEnvSchema]


class CondaPackage:
    def __init__(self, meta_json: Dict, prefix: Path):
        self.prefix = prefix
        self.name: str = meta_json["name"]
        self.version: str = meta_json["version"]
        self.subdir: str = meta_json["subdir"]
        self.files: str = meta_json["files"]
        self.depends: List[str] = meta_json.get("depends", [])
        self.constrains: List[str] = meta_json.get("constrains", [])
        if meta_json["channel"] == "<unknown>":
            logger.warning(f"Channel for {self.name} is unknown, setting to conda-forge")
            meta_json["channel"] = "conda-forge"
        channel_regex = rf"(.*\.\w*)/?(.*)/{self.subdir}$"
        result = re.match(channel_regex, meta_json["channel"])
        if not result:
            self.channel_url = f"https://conda.anaconda.org/{meta_json['channel']}"
            self.channel: str = meta_json["channel"]
        else:
            self.channel_url = result.group(1) + "/" + result.group(2)
            self.channel: str = result.group(2)

    def __repr__(self):
        return (
            f"CondaPackage(meta_json={{'name': {self.name!r}, 'version': "
            f"{self.version!r}, 'subdir': {self.subdir!r}, 'files': {self.files!r}, "
            f"'depends': {self.depends!r}, 'constrains': {self.constrains!r}, "
            f"'channel': {self.channel!r}}}, prefix={self.prefix!r})"
        )

    def __str__(self):
        return f"{self.name} {self.version} from {self.channel_url}"


class PackageLevel(TypedDict):
    name: str
    level: PackageLevelEnum
    source: Literal["pip", "conda"]


class ApiBase(TypedDict):
    id: int
    created: str
    updated: str


class SoftwareEnvironmentBuild(ApiBase):
    state: Literal["built", "building", "error", "queued"]


class SoftwareEnvironmentSpec(ApiBase):
    latest_build: Optional[SoftwareEnvironmentBuild]


class SoftwareEnvironmentAlias(ApiBase):
    latest_spec: Optional[SoftwareEnvironmentSpec]


class ArchitectureTypesEnum(str, Enum):
    """
    All currently supported architecture types
    """

    X86_64 = "x86_64"
    ARM64 = "aarch64"

    def __str__(self) -> str:
        return self.value

    @property
    def conda_suffix(self) -> str:
        if self == ArchitectureTypesEnum.X86_64:
            return "64"
        else:
            return self.value

    @property
    def vm_arch(self) -> Literal["x86_64", "arm64"]:
        if self == ArchitectureTypesEnum.ARM64:
            return "arm64"
        else:
            return self.value


class ClusterDetailsState(TypedDict):
    state: str
    reason: str
    updated: str


class ClusterDetailsProcess(TypedDict):
    created: str
    name: str
    current_state: ClusterDetailsState
    instance: dict


class ClusterDetails(TypedDict):
    id: int
    name: str
    workers: List[ClusterDetailsProcess]
    scheduler: Optional[ClusterDetailsProcess]
    current_state: ClusterDetailsState
    created: str


class FirewallOptions(TypedDict):
    """
    A dictionary with the following key/value pairs

    Parameters
    ----------
    ports
        List of ports to open to cidr on the scheduler.
        For example, ``[22, 8786]`` opens port 22 for SSH and 8786 for client to Dask connection.
    cidr
        CIDR block from which to allow access. For example ``0.0.0.0/0`` allows access from any IP address.
    """

    ports: List[int]
    cidr: str


class BackendOptions(TypedDict, total=False):
    """
    A dictionary with the following key/value pairs

    Parameters
    ----------
    region_name
        Region name to launch cluster in. For example: us-east-2
    zone_name
        Zone name to launch cluster in. For example: us-east-2a
    firewall
        Deprecated; use ``ingress`` instead.
    ingress
        Allows you to specify multiple CIDR blocks (and corresponding ports) to open for ingress
        on the scheduler firewall.
    spot
        Whether to request spot instances.
    spot_on_demand_fallback
        If requesting spot, whether to request non-spot instances if we get fewer spot instances
        than desired.
    spot_replacement
        By default we'll attempt to replace interrupted spot instances; set to False to disable.
    multizone
        Tell the cloud provider to pick zone with best availability, we'll keep workers all in the
        same zone, scheduler may or may not be in that zone as well.
    use_dashboard_public_ip
        Public IP is used by default, lets you choose to use private IP for dashboard link.
    use_dashboard_https
        When public IP address is used for dashboard, we'll enable HTTPS + auth by default.
        You may want to disable this if using something that needs to connect directly to
        the scheduler dashboard without authentication, such as jupyter dask-labextension.
    network_volumes
        Very experimental option to allow mounting SMB volume on cluster nodes.
    """

    region_name: Optional[str]
    zone_name: Optional[str]
    firewall: Optional[FirewallOptions]  # TODO deprecate, use ingress instead
    ingress: Optional[List[FirewallOptions]]
    spot: Optional[bool]
    spot_on_demand_fallback: Optional[bool]
    spot_replacement: Optional[bool]
    multizone: Optional[bool]
    use_dashboard_public_ip: Optional[bool]
    use_dashboard_https: Optional[bool]
    send_prometheus_metrics: Optional[bool]  # TODO deprecate
    prometheus_write: Optional[dict]  # TODO deprecate
    network_volumes: Optional[List[dict]]


class AWSOptions(BackendOptions, total=False):
    """
    A dictionary with the following key/value pairs plus any pairs in :py:class:`BackendOptions`

    Parameters
    ----------
    keypair_name
        AWS Keypair to assign worker/scheduler instances. This would need to be an existing keypair in your
            account, and needs to be in the same region as your cluster. Note that Coiled can also manage
            adding a unique, ephemeral keypair for SSH access to your cluster;
            see :doc:`ssh` for more information.
    use_placement_group
        If possible, this will attempt to put workers in the same cluster placement group (in theory this can
        result in better network between workers, since they'd be physically close to each other in datacenter,
        though we haven't seen this to have much benefit in practice).
    """

    keypair_name: Optional[str]
    use_placement_group: Optional[bool]


class GCPOptions(BackendOptions, total=False):
    """
    A dictionary with GCP specific key/value pairs plus any pairs in :py:class:`BackendOptions`
    """

    scheduler_accelerator_count: Optional[int]
    scheduler_accelerator_type: Optional[str]
    worker_accelerator_count: Optional[int]
    worker_accelerator_type: Optional[str]



from .cluster import Cluster
from .core import (
    CloudV2,
    cluster_details,
    cluster_logs,
    create_cluster,
    delete_cluster,
    list_clusters,
    log_cluster_debug_info,
    setup_logging,
    better_cluster_logs,
)
from ..types import AWSOptions, GCPOptions, BackendOptions, FirewallOptions

__all__ = [
    "AWSOptions",
    "GCPOptions",
    "FirewallOptions",
    "BackendOptions",
    "CloudV2",
    "cluster_details",
    "cluster_logs",
    "Cluster",
    "create_cluster",
    "delete_cluster",
    "list_clusters",
    "log_cluster_debug_info",
    "setup_logging",
]



import asyncio
import logging
import sys
from csv import writer as csv_writer
from logging import basicConfig
from pathlib import Path

import click
from rich.console import Console
from rich.table import Table

from coiled.scan import scan_prefix
from coiled.utils import COILED_LOCAL_PACKAGE_PREFIX, IGNORE_PYTHON_DIRS, recurse_importable_python_files


@click.group()
def package_sync():
    basicConfig(level=logging.INFO)


@package_sync.command()
@click.option("--csv", is_flag=True, default=False, help="Output as CSV")
@click.option("--verbose", "-v", is_flag=True, default=False, help="Output files that will end up in wheels")
def scan(csv: bool, verbose: bool):
    result = asyncio.run(scan_prefix())
    table = Table(title="Packages")
    table.add_column("Package Name", style="cyan", no_wrap=True)
    table.add_column("Conda Name", style="cyan", no_wrap=True)
    table.add_column("Version", style="magenta")
    table.add_column("Source", style="magenta")
    table.add_column("Channel", style="magenta", overflow="fold")
    table.add_column("Channel URL", style="magenta", overflow="fold")
    table.add_column("Wheel target", style="green", overflow="fold")
    table.add_column("Path", overflow="fold")
    if verbose:
        table.add_column("Python modules", overflow="fold")
    rows = []
    for pkg in result:
        row = [
            pkg["name"],
            pkg["conda_name"],
            pkg["version"],
            pkg["source"],
            pkg["channel"] or "",
            pkg["channel_url"] or "",
            pkg["wheel_target"] or "",
            str(pkg["path"]) if pkg["path"] else "",
        ]
        if verbose:
            if pkg["name"].startswith(COILED_LOCAL_PACKAGE_PREFIX) and pkg["path"]:
                modules = []
                for path in recurse_importable_python_files(pkg["path"]):
                    parent_names = {parent.name for parent in path.parents}
                    if not parent_names.intersection(IGNORE_PYTHON_DIRS) and str(path) not in (
                        "__init__.py",
                        "__main__.py",
                    ):
                        modules.append(str(path))
                row.append("\n".join(modules))
            else:
                row.append("")

        rows.append(row)

    rows = sorted(rows, key=lambda x: (x[3], x[0].lower()))

    if csv:
        writer = csv_writer(sys.stdout, lineterminator="\n")
        writer.writerow([c.header for c in table.columns])
        writer.writerows(rows)
    else:
        for row in rows:
            table.add_row(*row)
        console = Console()
        console.print(table)


@package_sync.command()
@click.option("--csv", is_flag=True, default=False, help="Output as CSV")
def debug(csv: bool):
    table = Table(title="Debug")
    table.add_column("Path", no_wrap=True, overflow="fold")
    rows = []
    for path in sys.path:
        p = Path(path)
        if p.is_dir():
            for file in p.iterdir():
                rows.append([str(file)])
        else:
            rows.append([str(p)])
    rows = sorted(rows)
    if csv:
        writer = csv_writer(sys.stdout, lineterminator="\n")
        writer.writerow([c.header for c in table.columns])
        writer.writerows(rows)
    else:
        for row in rows:
            table.add_row(*row)
        console = Console()
        console.print(table)



import asyncio

import click

from ..utils import CONTEXT_SETTINGS
from .amp import aws_amp_setup
from .aws import aws_setup
from .entry import do_setup_wizard, setup_wizard
from .gcp import gcp_setup
from .prometheus import get_prometheus_read_creds


@click.group(context_settings=CONTEXT_SETTINGS, invoke_without_command=True)
@click.pass_context
def setup(ctx):
    """Setup Coiled with cloud provider"""
    # `coiled setup` by itself should invoke wizard
    if ctx.invoked_subcommand is None:
        asyncio.run(do_setup_wizard())


setup.add_command(setup_wizard, "wizard")
setup.add_command(aws_setup, "aws")
setup.add_command(gcp_setup, "gcp")

setup.add_command(aws_amp_setup, "amp")
setup.add_command(get_prometheus_read_creds, "prometheus-datasource")



from __future__ import annotations

from collections import Counter
from contextlib import contextmanager
from typing import Any, Mapping, Optional

import jmespath
from rich.progress import Progress
from typing_extensions import Literal

from ..states import CombinedProcessStateEnum, get_combined_process_state

SCHEDULER_SEARCH = jmespath.compile("scheduler.instance.instance_type_id")
WORKER_SEARCH = jmespath.compile("workers[*].instance")


def get_worker_statuses(cluster_details) -> Counter[CombinedProcessStateEnum]:
    """Get worker status from a cluster_details response.

    Returns a Counter of the worker statuses listed in `CombinedProcessStateEnum`.
    """
    worker_statuses: Counter[CombinedProcessStateEnum] = Counter(
        get_combined_process_state(worker) for worker in cluster_details["workers"]
    )
    return worker_statuses


def get_instance_types(cluster_details: Mapping[str, Any]) -> tuple[str | None, Counter[str | None]]:
    """Get instance types from a cluster_details response.

    Returns a tuple of schduler_instance_type, Counter(worker_instance_types).
    If the instance type is still not determined, will show "Unknown".
    """
    scheduler_instance_type = SCHEDULER_SEARCH.search(cluster_details) or None
    worker_instance_types = Counter(w.get("instance_type_id", None) for w in WORKER_SEARCH.search(cluster_details))
    return scheduler_instance_type, worker_instance_types


def sniff_environment() -> Literal["notebook", "ipython_terminal", "terminal"]:
    """Heuristically determine the execution context.

    This function attempts to determine whether the execution environment is a Jupyter
    Notebook-like context, an IPython REPL context, or the standard Python context.
    The detection is not foolproof and is sensitive to the implementation details of
    the shell.
    """
    try:
        get_ipython  # type: ignore
    except NameError:
        return "terminal"
    ipython = get_ipython()  # type: ignore # noqa: F821
    shell = ipython.__class__.__name__
    if "google.colab" in str(ipython.__class__) or shell == "ZMQInteractiveShell":
        return "notebook"  # Jupyter notebook or qtconsole
    elif shell == "TerminalInteractiveShell":
        return "ipython_terminal"  # Terminal running IPython
    else:
        return "terminal"  # Other type (?)


@contextmanager
def simple_progress(description: str, progress: Optional[Progress] = None):
    if progress:
        task = progress.add_task(description=description, total=1)
        yield
        progress.advance(task)
    else:
        yield



from typing import Any, Mapping, Optional

from typing_extensions import Protocol


class ClusterWidget(Protocol):
    def update(
        self,
        cluster_details: Optional[Mapping[str, Any]],
        logs,
        *args,
        final_update=None,
        **kwargs,
    ) -> None:
        pass



import sys

import distributed
from importlib_metadata import version
from packaging.version import Version

COILED_VERSION = version("coiled")

PY_VERSION = Version(".".join(map(str, sys.version_info[:3])))

DISTRIBUTED_VERSION = Version(distributed.__version__)


def register_plugin(client, plugin, **kwargs):
    if DISTRIBUTED_VERSION >= Version("2023.9.2"):
        return client.register_plugin(plugin, **kwargs)
    elif isinstance(plugin, distributed.SchedulerPlugin):
        return client.register_scheduler_plugin(plugin, **kwargs)
    elif isinstance(plugin, (distributed.WorkerPlugin, distributed.NannyPlugin)):
        return client.register_worker_plugin(plugin, **kwargs)
    else:
        raise TypeError(f"Invalid plugin type {type(plugin)}")



from json import loads as json_loads

import click
from rich import print

import coiled

from .utils import CONTEXT_SETTINGS


@click.command(
    context_settings=CONTEXT_SETTINGS,
    help="CLI to hit endpoints using Coiled account authentication (mostly for internal use)",
)
@click.argument("url")
@click.option("-X", "--request", default="GET")
@click.option("-d", "--data", multiple=True)
@click.option("--json", is_flag=True, default=False)
@click.option("--json-output", is_flag=True, default=False, help="Set to pretty print JSON responses")
def curl(url: str, request, data, json, json_output):
    all_data = "&".join(data) if data else None
    with coiled.Cloud() as cloud:
        if "http" not in url:
            url = f"{cloud.server}{url}"

        response = sync_request(cloud, url, method=request, data=all_data, json=json, json_output=json_output)

    print(response)


def sync_request(cloud, url, method, data, json: bool = False, json_output: bool = False):
    kwargs = {"method": method, "url": url}

    if json:
        kwargs["json"] = json_loads(data)
    else:
        kwargs["data"] = data
    response = cloud._sync(cloud._do_request, **kwargs)
    if response.status >= 400:
        print(f"{url} returned {response.status}")

    async def get_result(r):
        return await (r.json() if json_output else r.text())

    return cloud._sync(
        get_result,
        response,
    )



from __future__ import annotations

import os
import sys
import threading
import warnings
from collections.abc import Iterator
from typing import Dict, Optional, Union

import dask.config
import dask.distributed
import toolz
from dask.base import tokenize
from dask.utils import parse_timedelta

import coiled
from coiled.utils import error_info_for_tracking

_clients = {}
_lock = threading.RLock()
_clusters = {}


class Function:
    """A function that you can run remotely"""

    def __init__(self, function, cluster_kwargs, keepalive, environ):
        self.function = function
        self._cluster_kwargs = cluster_kwargs
        self.keepalive = parse_timedelta(keepalive)
        self._environ = environ

        token = tokenize(
            sys.executable,
            # TODO: include something about the software environment
            **cluster_kwargs,
        )
        self._name = f"function-{token[:8]}"

    @property
    def cluster(self) -> coiled.Cluster:
        with _lock:
            try:
                return _clusters[self._name]
            except KeyError:
                success = True
                exception = None
                info = {}
                info["keepalive"] = self.keepalive
                try:
                    # Setting to use the local threaded scheduler avoids implicit tasks in tasks.
                    # This relies on `send_dask_config=True` (default value).
                    with dask.config.set({"scheduler": "threads", "distributed.worker.daemon": False}):
                        cluster = coiled.Cluster(name=self._name, **self._cluster_kwargs)
                        info["account"] = cluster.account
                        info["cluster_id"] = cluster.cluster_id
                    if self._environ:
                        cluster.send_private_envs(self._environ)
                    cluster.adapt(minimum=0, maximum=500)
                    _clusters[self._name] = cluster
                    return cluster
                except Exception as e:
                    success = False
                    exception = e
                    raise
                finally:
                    coiled.add_interaction(
                        "coiled-function",
                        success=success,
                        **info,
                        **error_info_for_tracking(exception),
                    )

    @property
    def client(self) -> dask.distributed.Client:
        with _lock:
            try:
                return _clients[self._name]
            except KeyError:
                client = dask.distributed.Client(self.cluster, set_as_default=False)
                if self.cluster.shutdown_on_close:
                    self.cluster.set_keepalive(keepalive=self.keepalive)
                _clients[self._name] = client
                return client

    def __call__(self, *args, **kwargs):
        # If this is being called from on the desired cluster, then run locally.
        # This allows one Function (with same desired cluster specs) to call another without overhead.
        if os.environ.get("COILED_CLUSTER_NAME", None) == self._name:
            return self.local(*args, **kwargs)
        # Otherwise, submit to cluster.
        return self.submit(*args, **kwargs).result()

    def local(self, *args, **kwargs):
        return self.function(*args, **kwargs)

    def submit(self, *args, **kwargs) -> dask.distributed.Future:
        """Submit function call for asynchronous execution

        This immediately returns a Dask Future, allowing for the submission of
        many tasks in parallel.

        Example
        -------
        >>> @coiled.function()
        ... def f(x):
        ...    return x + 1

        >>> f(10)  # calling the function blocks until finished
        11
        >>> f.submit(10)  # immediately returns a future
        <Future: pending, key=f-1234>
        >>> f.submit(10).result()  # Call .result to get result
        11

        >>> futures = [f(i) for i in range(1000)]  # parallelize with a for loop
        >>> [future.result() for future in futures]
        ...

        Returns
        -------
        future: dask.distributed.Future

        See Also
        --------
        Function.map
        """
        return self.client.submit(self.function, *args, **kwargs)

    def map(self, *args, **kwargs) -> Iterator:
        """Map function across many inputs

        This runs your function many times in parallel across all of the items
        in an input list.  Coiled will auto-scale your cluster to meet demand.

        Example
        -------
        >>> @coiled.function()
        ... def process(filename: str):
        ...     " Convert CSV file to Parquet "
        ...     df = pd.read_csv(filename)
        ...     outfile = filename[:-4] + ".parquet"
        ...     df.to_parquet(outfile)
        ...     return outfile

        >>> process("s3://my-bucket/data.csv")  # calling the function blocks until finished
        11
        >>> filenames = process.map(filenames)
        >>> print(list(filenames))  # print out all output filenames

        Returns
        -------
        results: Iterator

        See Also
        --------
        Function.submit
        """
        kwargs.setdefault("pure", False)
        kwargs.setdefault("batch_size", 100)
        futures = self.client.map(self.function, *args, **kwargs)
        batchsize = max(int(len(futures) / 50), 1)  # type: ignore
        batches = toolz.partition_all(batchsize, futures)
        return (result for batch in batches for result in self.client.gather(batch))  # type: ignore


def function(
    *,
    software: Optional[str] = None,
    container: Optional[str] = None,
    vm_type: Optional[Union[str, list[str]]] = None,
    cpu: Optional[Union[int, list[int]]] = None,
    memory: Optional[Union[str, list[str]]] = None,
    gpu: Optional[bool] = None,
    account: Optional[str] = None,
    region: Optional[str] = None,
    arm: Optional[bool] = None,
    disk_size: Optional[int] = None,
    shutdown_on_close: bool = True,
    spot_policy: Optional[str] = None,
    idle_timeout: str = "24 hours",
    keepalive="30 seconds",
    environ: Optional[Dict[str, str]] = None,
    threads_per_worker: Union[int, None] = 1,
):
    """
    Decorate a function to run on cloud infrastructure

    This creates a ``Function`` object that executes its code on a remote cluster
    with the hardware and software specified in the arguments to the decorator.
    It can run either as a normal function, or it can return Dask Futures for
    parallel computing.

    Parameters
    ----------
    software
        Name of the software environment to use; this allows you to use and re-use existing
        Coiled software environments, and should not be used with package sync or when specifying
        a container to use for this specific cluster.
    container
        Name or URI of container image to use; when using a pre-made container image with Coiled,
        this allows you to skip the step of explicitly creating a Coiled software environment
        from that image. Note that this should not be used with package sync or when specifying
        an existing Coiled software environment.
    vm_type
        Instance type, or list of instance types, that you would like to use.
        You can use ``coiled.list_instance_types()`` to see a list of allowed types.
    cpu
        Number, or range, of CPUs requested. Specify a range by
        using a list of two elements, for example: ``cpu=[2, 8]``.
    memory
        Amount of memory to request for each VM, Coiled will use a +/- 10% buffer
        from the memory that you specify. You may specify a range of memory by using a
        list of two elements, for example: ``memory=["2GiB", "4GiB"]``.
    disk_size
        Size of persistent disk attached to each VM instance, specified in GiB.
    gpu
        Whether to attach a GPU; this would be a single NVIDIA T4.
    region
        The cloud provider region in which to run the cluster.
    arm
        Whether to use ARM instances for cluster; default is x86 (Intel) instances.
    keepalive
        Keep your cluster running for the specified time, even if your Python session closes.
        Default is "30 seconds".
    spot_policy
        Purchase option to use for workers in your cluster, options are "on-demand", "spot", and
        "spot_with_fallback"; by default this is "on-demand".
        (Google Cloud refers to this as "provisioning model" for your instances.)
        **Spot instances** are much cheaper, but can have more limited availability and may be terminated
        while you're still using them if the cloud provider needs more capacity for other customers.
        **On-demand instances** have the best availability and are almost never
        terminated while still in use, but they're significantly more expensive than spot instances.
        For most workloads, "spot_with_fallback" is likely to be a good choice: Coiled will try to get as
        many spot instances as we can, and if we get less than you requested, we'll try to get the remaining
        instances as on-demand.
        For AWS, when we're notified that an active spot instance is going to be terminated,
        we'll attempt to get a replacement instance (spot if available, but could be on-demand if you've
        enabled "fallback"). Dask on the active instance will attempt a graceful shutdown before the
        instance is terminated so that computed results won't be lost.
    idle_timeout
        Shut down the cluster after this duration if no activity has occurred. Default is "24 hours".
    environ
        Dictionary of environment variables to securely pass to the cloud VM environment.
    threads_per_worker
        Number of threads to run concurrent tasks in for each VM. -1 can be used to run as many concurrent
        tasks as there are CPU cores. Default is 1.


    See the :class:`coiled.Cluster` docstring for additional parameter descriptions.

    Examples
    --------
    >>> import coiled
    >>> @coiled.function()
    ... def f(x):
    ...    return x + 1

    >>> f(10)  # calling the function blocks until finished
    11
    >>> f.submit(10)  # immediately returns a future
    <Future: pending, key=f-1234>
    >>> f.submit(10).result()  # Call .result to get result
    11

    >>> futures = [f(i) for i in range(1000)]  # parallelize with a for loop
    >>> [future.result() for future in futures]
    ...
    """

    def decorator(func) -> Function:
        nonlocal cpu, threads_per_worker, environ

        default_environ = coiled.utils.unset_single_thread_defaults()
        if container and "rapidsai" in container:
            default_environ = {"DISABLE_JUPYTER": "true", **default_environ}  # needed for "stable" RAPIDS image

        if memory is None and cpu is None and not vm_type:
            cpu = 2

        if threads_per_worker == -1:
            # Have `-1` mean the same as CPU count (Dask's default behavior)
            threads_per_worker = None

        cluster_kwargs = dict(
            account=account,
            n_workers=0,
            scheduler_cpu=cpu,
            scheduler_memory=memory,
            worker_cpu=cpu,
            worker_memory=memory,
            software=software,
            container=container,
            idle_timeout=idle_timeout,
            scheduler_vm_types=vm_type,
            worker_vm_types=vm_type,
            allow_ssh=True,
            environ=default_environ,
            scheduler_gpu=gpu,
            region=region,
            arm=arm,
            shutdown_on_close=shutdown_on_close,
            spot_policy=spot_policy,
            extra_worker_on_scheduler=True,
            tags={"coiled-cluster-type": "function"},
            worker_options={"nthreads": threads_per_worker},
            scheduler_disk_size=disk_size,
            worker_disk_size=disk_size,
        )

        return Function(func, cluster_kwargs, keepalive=keepalive, environ=environ)

    return decorator


# Small backwards compatibility shim
def run(*args, **kwargs):
    warnings.warn(
        "coiled.run has been renamed to coiled.function. "
        "Please use coiled.function as coiled.run will be removed in a future release.",
        FutureWarning,
        stacklevel=2,
    )
    return function(*args, **kwargs)
